import { Map, OrderedSet } from "immutable";
import { Type, NamedType, ClassType, EnumType, UnionType, TypeKind } from "./Type";
import { Namespace, Name, Namer, DependencyName } from "./Naming";
import { Renderer, BlankLineLocations } from "./Renderer";
import { Sourcelike } from "./Source";
export declare abstract class ConvenienceRenderer extends Renderer {
    protected globalNamespace: Namespace;
    private _topLevelNames;
    private _namesForNamedTypes;
    private _propertyNames;
    private _memberNames;
    private _caseNames;
    private _namedTypes;
    private _namedClasses;
    private _namedEnums;
    private _namedUnions;
    private _haveUnions;
    private _alphabetizeProperties;
    readonly topLevels: Map<string, Type>;
    protected readonly forbiddenNamesForGlobalNamespace: string[];
    protected forbiddenForClassProperties(_c: ClassType, _className: Name): {
        names: Name[];
        namespaces: Namespace[];
    };
    protected forbiddenForUnionMembers(_u: UnionType, _unionName: Name): {
        names: Name[];
        namespaces: Namespace[];
    };
    protected forbiddenForEnumCases(_e: EnumType, _enumName: Name): {
        names: Name[];
        namespaces: Namespace[];
    };
    protected topLevelDependencyNames(_t: Type, _topLevelName: Name): DependencyName[];
    protected namedTypeDependencyNames(_t: NamedType, _name: Name): DependencyName[];
    protected abstract topLevelNameStyle(rawName: string): string;
    protected readonly abstract namedTypeNamer: Namer;
    protected readonly abstract classPropertyNamer: Namer | null;
    protected readonly abstract unionMemberNamer: Namer | null;
    protected readonly abstract enumCaseNamer: Namer | null;
    protected abstract namedTypeToNameForTopLevel(type: Type): NamedType | null;
    protected abstract emitSourceStructure(): void;
    protected readonly unionMembersInGlobalNamespace: boolean;
    protected readonly enumCasesInGlobalNamespace: boolean;
    protected unionNeedsName(u: UnionType): boolean;
    protected setUpNaming(): Namespace[];
    private addDependenciesForNamedType;
    private nameForTopLevel;
    private addNamedForNamedType;
    private addPropertyNames;
    private makeUnionMemberName(u, unionName, t);
    private addUnionMemberNames;
    private addEnumCaseNames;
    private childrenOfType;
    protected readonly namedUnions: OrderedSet<UnionType>;
    protected readonly haveNamedUnions: boolean;
    protected readonly haveUnions: boolean;
    protected readonly enums: OrderedSet<EnumType>;
    protected readonly haveEnums: boolean;
    protected proposedUnionMemberNameForTypeKind: (_kind: TypeKind) => string | null;
    protected proposeUnionMemberName(_u: UnionType, _unionName: Name, fieldType: Type, lookup: (n: Name) => string): string;
    protected nameForNamedType: (t: NamedType) => Name;
    protected forEachTopLevel: (blankLocations: BlankLineLocations, f: (t: Type, name: Name) => void, predicate?: ((t: Type) => boolean) | undefined) => void;
    setAlphabetizeProperties: (value: boolean) => void;
    protected forEachClassProperty: (c: ClassType, blankLocations: BlankLineLocations, f: (name: Name, jsonName: string, t: Type) => void) => void;
    protected nameForUnionMember: (u: UnionType, t: Type) => Name;
    protected forEachUnionMember: (u: UnionType, members: OrderedSet<Type> | null, blankLocations: BlankLineLocations, sortOrder: ((n: Name, t: Type) => string) | null, f: (name: Name, t: Type) => void) => void;
    protected forEachEnumCase: (e: EnumType, blankLocations: BlankLineLocations, f: (name: Name, jsonName: string) => void) => void;
    protected callForNamedType<T extends NamedType>(t: T, f: (t: T, name: Name) => void): void;
    protected forEachSpecificNamedType<T extends NamedType>(blankLocations: BlankLineLocations, types: OrderedSet<T>, f: (t: T, name: Name) => void): void;
    protected forEachClass: (blankLocations: BlankLineLocations, f: (c: ClassType, className: Name) => void) => void;
    protected forEachEnum: (blankLocations: BlankLineLocations, f: (u: EnumType, enumName: Name) => void) => void;
    protected forEachUnion: (blankLocations: BlankLineLocations, f: (u: UnionType, unionName: Name) => void) => void;
    protected forEachUniqueUnion<T>(blankLocations: BlankLineLocations, uniqueValue: (u: UnionType) => T, f: (firstUnion: UnionType, value: T) => void): void;
    protected forEachNamedType: (blankLocations: BlankLineLocations, leavesFirst: boolean, classFunc: (c: ClassType, className: Name) => void, enumFunc: (e: EnumType, enumName: Name) => void, unionFunc: (u: UnionType, unionName: Name) => void) => void;
    protected sourcelikeToString: (src: Sourcelike) => string;
    protected emitSource(): void;
}
