"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Naming_1 = require("./Naming");
var Renderer_1 = require("./Renderer");
var Support_1 = require("./Support");
var Source_1 = require("./Source");
var ConvenienceRenderer = /** @class */ (function (_super) {
    __extends(ConvenienceRenderer, _super);
    function ConvenienceRenderer() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this._alphabetizeProperties = false;
        _this.addDependenciesForNamedType = function (type, named) {
            var dependencyNames = _this.namedTypeDependencyNames(type, named);
            for (var _i = 0, dependencyNames_1 = dependencyNames; _i < dependencyNames_1.length; _i++) {
                var dn = dependencyNames_1[_i];
                _this.globalNamespace.add(dn);
            }
        };
        _this.nameForTopLevel = function (type, name) {
            var maybeNamedType = _this.namedTypeToNameForTopLevel(type);
            var styledName;
            if (maybeNamedType) {
                styledName = _this.namedTypeNamer.nameStyle(name);
            }
            else {
                styledName = _this.topLevelNameStyle(name);
            }
            var named = _this.globalNamespace.add(new Naming_1.FixedName(styledName));
            var dependencyNames = _this.topLevelDependencyNames(type, named);
            for (var _i = 0, dependencyNames_2 = dependencyNames; _i < dependencyNames_2.length; _i++) {
                var dn = dependencyNames_2[_i];
                _this.globalNamespace.add(dn);
            }
            if (maybeNamedType) {
                _this.addDependenciesForNamedType(maybeNamedType, named);
                _this._namesForNamedTypes = _this._namesForNamedTypes.set(maybeNamedType, named);
            }
            return named;
        };
        _this.addNamedForNamedType = function (type) {
            var existing = _this._namesForNamedTypes.get(type);
            if (existing !== undefined)
                return existing;
            var named = _this.globalNamespace.add(new Naming_1.SimpleName(type.proposedNames, _this.namedTypeNamer));
            _this.addDependenciesForNamedType(type, named);
            _this._namesForNamedTypes = _this._namesForNamedTypes.set(type, named);
            return named;
        };
        _this.addPropertyNames = function (c, className) {
            var propertyNamer = _this.classPropertyNamer;
            if (propertyNamer === null)
                return;
            var _a = _this.forbiddenForClassProperties(c, className), forbiddenNames = _a.names, forbiddenNamespace = _a.namespaces;
            var ns = new Naming_1.Namespace(c.combinedName, _this.globalNamespace, immutable_1.Set(forbiddenNamespace), immutable_1.Set(forbiddenNames));
            var names = c.sortedProperties
                .map(function (_, name) {
                // FIXME: This alternative should really depend on what the
                // actual name of the class ends up being.  We can do this
                // with a DependencyName.
                // Also, we currently don't have any languages where properties
                // are global, so collisions here could only occur where two
                // properties of the same class have the same name, in which case
                // the alternative would also be the same, i.e. useless.  But
                // maybe we'll need global properties for some weird language at
                // some point.
                var alternative = c.combinedName + "_" + name;
                return ns.add(new Naming_1.SimpleName(immutable_1.OrderedSet([name, alternative]), propertyNamer));
            })
                .toMap();
            _this._propertyNames = _this._propertyNames.set(c, names);
        };
        _this.addUnionMemberNames = function (u, unionName) {
            var memberNamer = _this.unionMemberNamer;
            if (memberNamer === null)
                return;
            var _a = _this.forbiddenForUnionMembers(u, unionName), forbiddenNames = _a.names, forbiddenNamespace = _a.namespaces;
            var ns;
            if (_this.unionMembersInGlobalNamespace) {
                ns = _this.globalNamespace;
            }
            else {
                ns = new Naming_1.Namespace(u.combinedName, _this.globalNamespace, immutable_1.Set(forbiddenNamespace), immutable_1.Set(forbiddenNames));
            }
            var names = immutable_1.Map();
            u.members.forEach(function (t) {
                var name = _this.makeUnionMemberName(u, unionName, t);
                names = names.set(t, ns.add(name));
            });
            _this._memberNames = _this._memberNames.set(u, names);
        };
        // FIXME: this is very similar to addPropertyNameds and addUnionMemberNames
        _this.addEnumCaseNames = function (e, enumName) {
            var caseNamer = _this.enumCaseNamer;
            if (caseNamer === null)
                return;
            var _a = _this.forbiddenForEnumCases(e, enumName), forbiddenNames = _a.names, forbiddenNamespace = _a.namespaces;
            var ns;
            if (_this.enumCasesInGlobalNamespace) {
                ns = _this.globalNamespace;
            }
            else {
                ns = new Naming_1.Namespace(e.combinedName, _this.globalNamespace, immutable_1.Set(forbiddenNamespace), immutable_1.Set(forbiddenNames));
            }
            var names = immutable_1.Map();
            e.cases.forEach(function (name) {
                // FIXME: See the FIXME in `addPropertyNameds`.  We do have global
                // enum cases, though (in Go), so this is actually useful already.
                var alternative = e.combinedName + "_" + name;
                names = names.set(name, ns.add(new Naming_1.SimpleName(immutable_1.OrderedSet([name, alternative]), caseNamer)));
            });
            _this._caseNames = _this._caseNames.set(e, names);
        };
        _this.childrenOfType = function (t) {
            var names = _this.names;
            if (t instanceof Type_1.ClassType && _this.classPropertyNamer !== null) {
                var propertyNameds_1 = Support_1.defined(_this._propertyNames.get(t));
                return t.properties
                    .sortBy(function (_, n) { return Support_1.defined(names.get(Support_1.defined(propertyNameds_1.get(n)))); })
                    .toOrderedSet();
            }
            return t.children.toOrderedSet();
        };
        _this.proposedUnionMemberNameForTypeKind = function (_kind) {
            return null;
        };
        _this.nameForNamedType = function (t) {
            var name = _this._namesForNamedTypes.get(t);
            if (name === undefined) {
                return Support_1.panic("Named type does not exist.");
            }
            return name;
        };
        _this.forEachTopLevel = function (blankLocations, f, predicate) {
            var topLevels;
            if (predicate) {
                topLevels = _this.topLevels.filter(predicate);
            }
            else {
                topLevels = _this.topLevels;
            }
            _this.forEachWithBlankLines(topLevels, blankLocations, function (t, name) {
                return f(t, Support_1.defined(_this._topLevelNames.get(name)));
            });
        };
        _this.setAlphabetizeProperties = function (value) {
            _this._alphabetizeProperties = value;
        };
        _this.forEachClassProperty = function (c, blankLocations, f) {
            var propertyNames = Support_1.defined(_this._propertyNames.get(c));
            if (_this._alphabetizeProperties) {
                var alphabetizedPropertyNames = propertyNames.sortBy(function (n) { return _this.names.get(n); }).toOrderedMap();
                _this.forEachWithBlankLines(alphabetizedPropertyNames, blankLocations, function (name, jsonName) {
                    var t = Support_1.defined(c.properties.get(jsonName));
                    f(name, jsonName, t);
                });
            }
            else {
                _this.forEachWithBlankLines(c.properties, blankLocations, function (t, jsonName) {
                    var name = Support_1.defined(propertyNames.get(jsonName));
                    f(name, jsonName, t);
                });
            }
        };
        _this.nameForUnionMember = function (u, t) {
            return Support_1.defined(Support_1.defined(_this._memberNames.get(u)).get(t));
        };
        _this.forEachUnionMember = function (u, members, blankLocations, sortOrder, f) {
            var iterateMembers = members === null ? u.members : members;
            if (sortOrder === null) {
                sortOrder = function (n) { return Support_1.defined(_this.names.get(n)); };
            }
            var memberNames = Support_1.defined(_this._memberNames.get(u)).filter(function (_, t) { return iterateMembers.has(t); });
            var sortedMemberNames = memberNames.sortBy(sortOrder).toOrderedMap();
            _this.forEachWithBlankLines(sortedMemberNames, blankLocations, f);
        };
        _this.forEachEnumCase = function (e, blankLocations, f) {
            var caseNames = Support_1.defined(_this._caseNames.get(e));
            var sortedCaseNames = caseNames.sortBy(function (n) { return _this.names.get(n); }).toOrderedMap();
            _this.forEachWithBlankLines(sortedCaseNames, blankLocations, f);
        };
        _this.forEachClass = function (blankLocations, f) {
            _this.forEachSpecificNamedType(blankLocations, _this._namedClasses, f);
        };
        _this.forEachEnum = function (blankLocations, f) {
            _this.forEachSpecificNamedType(blankLocations, _this._namedEnums, f);
        };
        _this.forEachUnion = function (blankLocations, f) {
            _this.forEachSpecificNamedType(blankLocations, _this._namedUnions, f);
        };
        _this.forEachNamedType = function (blankLocations, leavesFirst, classFunc, enumFunc, unionFunc) {
            var collection = _this._namedTypes;
            if (leavesFirst)
                collection = collection.reverse();
            _this.forEachWithBlankLines(collection, blankLocations, function (t) {
                if (t instanceof Type_1.ClassType) {
                    _this.callForNamedType(t, classFunc);
                }
                else if (t instanceof Type_1.EnumType) {
                    _this.callForNamedType(t, enumFunc);
                }
                else if (t instanceof Type_1.UnionType) {
                    _this.callForNamedType(t, unionFunc);
                }
                else {
                    return Support_1.panic("Named type that's neither a class nor union");
                }
            });
        };
        // You should never have to use this to produce parts of your generated
        // code.  If you need to modify a Name, for example to change its casing,
        // use `modifySource`.
        _this.sourcelikeToString = function (src) {
            return Source_1.serializeRenderResult({
                rootSource: Source_1.sourcelikeToSource(src),
                names: _this.names
            }, 
            // FIXME: Use proper indentation.
            "").lines.join("\n");
        };
        return _this;
    }
    Object.defineProperty(ConvenienceRenderer.prototype, "topLevels", {
        get: function () {
            return this.typeGraph.topLevels;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ConvenienceRenderer.prototype, "forbiddenNamesForGlobalNamespace", {
        get: function () {
            return [];
        },
        enumerable: true,
        configurable: true
    });
    ConvenienceRenderer.prototype.forbiddenForClassProperties = function (_c, _className) {
        return { names: [], namespaces: [] };
    };
    ConvenienceRenderer.prototype.forbiddenForUnionMembers = function (_u, _unionName) {
        return { names: [], namespaces: [] };
    };
    ConvenienceRenderer.prototype.forbiddenForEnumCases = function (_e, _enumName) {
        return { names: [], namespaces: [] };
    };
    ConvenienceRenderer.prototype.topLevelDependencyNames = function (_t, _topLevelName) {
        return [];
    };
    ConvenienceRenderer.prototype.namedTypeDependencyNames = function (_t, _name) {
        return [];
    };
    Object.defineProperty(ConvenienceRenderer.prototype, "unionMembersInGlobalNamespace", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ConvenienceRenderer.prototype, "enumCasesInGlobalNamespace", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    ConvenienceRenderer.prototype.unionNeedsName = function (u) {
        return !Type_1.nullableFromUnion(u);
    };
    ConvenienceRenderer.prototype.setUpNaming = function () {
        var _this = this;
        this.globalNamespace = Naming_1.keywordNamespace("global", this.forbiddenNamesForGlobalNamespace);
        var _a = this.typeGraph.allNamedTypesSeparated(), classes = _a.classes, enums = _a.enums, unions = _a.unions;
        var namedUnions = unions.filter(function (u) { return _this.unionNeedsName(u); }).toOrderedSet();
        this._namesForNamedTypes = immutable_1.Map();
        this._propertyNames = immutable_1.Map();
        this._memberNames = immutable_1.Map();
        this._caseNames = immutable_1.Map();
        this._topLevelNames = this.topLevels.map(this.nameForTopLevel).toMap();
        classes.forEach(function (c) {
            var name = _this.addNamedForNamedType(c);
            _this.addPropertyNames(c, name);
        });
        enums.forEach(function (e) {
            var name = _this.addNamedForNamedType(e);
            _this.addEnumCaseNames(e, name);
        });
        namedUnions.forEach(function (u) {
            var name = _this.addNamedForNamedType(u);
            _this.addUnionMemberNames(u, name);
        });
        return [this.globalNamespace];
    };
    ConvenienceRenderer.prototype.makeUnionMemberName = function (u, unionName, t) {
        var _this = this;
        return new Naming_1.DependencyName(Support_1.nonNull(this.unionMemberNamer), function (lookup) {
            return _this.proposeUnionMemberName(u, unionName, t, lookup);
        });
    };
    Object.defineProperty(ConvenienceRenderer.prototype, "namedUnions", {
        get: function () {
            return this._namedUnions;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ConvenienceRenderer.prototype, "haveNamedUnions", {
        get: function () {
            return !this._namedUnions.isEmpty();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ConvenienceRenderer.prototype, "haveUnions", {
        get: function () {
            return this._haveUnions;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ConvenienceRenderer.prototype, "enums", {
        get: function () {
            return this._namedEnums;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ConvenienceRenderer.prototype, "haveEnums", {
        get: function () {
            return !this._namedEnums.isEmpty();
        },
        enumerable: true,
        configurable: true
    });
    ConvenienceRenderer.prototype.proposeUnionMemberName = function (_u, _unionName, fieldType, lookup) {
        var _this = this;
        var simpleName = this.proposedUnionMemberNameForTypeKind(fieldType.kind);
        if (simpleName !== null) {
            return simpleName;
        }
        var typeNameForUnionMember = function (t) {
            return Type_1.matchTypeExhaustive(t, function (_anyType) { return "anything"; }, function (_nullType) { return "null"; }, function (_boolType) { return "bool"; }, function (_integerType) { return "integer"; }, function (_doubleType) { return "double"; }, function (_stringType) { return "string"; }, function (arrayType) { return typeNameForUnionMember(arrayType.items) + "_array"; }, function (classType) { return lookup(_this.nameForNamedType(classType)); }, function (mapType) { return typeNameForUnionMember(mapType.values) + "_map"; }, function (_enumType) { return "enum"; }, function (_unionType) { return "union"; }, function (_dateType) { return "date"; }, function (_timeType) { return "time"; }, function (_dateTimeType) { return "date_time"; });
        };
        return typeNameForUnionMember(fieldType);
    };
    ConvenienceRenderer.prototype.callForNamedType = function (t, f) {
        f(t, Support_1.defined(this._namesForNamedTypes.get(t)));
    };
    ConvenienceRenderer.prototype.forEachSpecificNamedType = function (blankLocations, types, f) {
        var _this = this;
        this.forEachWithBlankLines(types, blankLocations, function (t) {
            _this.callForNamedType(t, f);
        });
    };
    ConvenienceRenderer.prototype.forEachUniqueUnion = function (blankLocations, uniqueValue, f) {
        var firstUnionByValue = immutable_1.OrderedMap();
        this._namedUnions.forEach(function (u) {
            var v = uniqueValue(u);
            if (!firstUnionByValue.has(v)) {
                firstUnionByValue = firstUnionByValue.set(v, u);
            }
        });
        this.forEachWithBlankLines(firstUnionByValue, blankLocations, f);
    };
    ConvenienceRenderer.prototype.emitSource = function () {
        var _this = this;
        var types = this.typeGraph.allNamedTypes(this.childrenOfType);
        this._haveUnions = types.some(function (t) { return t instanceof Type_1.UnionType; });
        this._namedTypes = types
            .filter(function (t) { return !(t instanceof Type_1.UnionType) || _this.unionNeedsName(t); })
            .toOrderedSet();
        var _a = Type_1.separateNamedTypes(this._namedTypes), classes = _a.classes, enums = _a.enums, unions = _a.unions;
        this._namedClasses = classes;
        this._namedEnums = enums;
        this._namedUnions = unions;
        this.emitSourceStructure();
    };
    return ConvenienceRenderer;
}(Renderer_1.Renderer));
exports.ConvenienceRenderer = ConvenienceRenderer;
