"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var pluralize = require("pluralize");
var CompressedJSON_1 = require("./CompressedJSON");
var Support_1 = require("./Support");
var TypeBuilder_1 = require("./TypeBuilder");
var DateTime_1 = require("./DateTime");
var MIN_LENGTH_FOR_ENUM = 10;
function forEachArrayInNestedValueArray(va, f) {
    if (va.length === 0) {
        return;
    }
    if (Array.isArray(va[0])) {
        for (var _i = 0, va_1 = va; _i < va_1.length; _i++) {
            var x = va_1[_i];
            forEachArrayInNestedValueArray(x, f);
        }
    }
    else {
        f(va);
    }
}
function forEachValueInNestedValueArray(va, f) {
    forEachArrayInNestedValueArray(va, function (a) { return a.forEach(f); });
}
var InferenceUnionBuilder = /** @class */ (function (_super) {
    __extends(InferenceUnionBuilder, _super);
    function InferenceUnionBuilder(typeBuilder, typeName, isInferred, _typeInference, _cjson) {
        var _this = _super.call(this, typeBuilder, typeName, isInferred) || this;
        _this._typeInference = _typeInference;
        _this._cjson = _cjson;
        _this.setNumValues = function (n) {
            if (_this._numValues !== undefined) {
                return Support_1.panic("Can only set number of values once");
            }
            _this._numValues = n;
        };
        return _this;
    }
    InferenceUnionBuilder.prototype.makeEnum = function (enumCases) {
        Support_1.assert(enumCases.length > 0);
        var numValues = Support_1.defined(this._numValues);
        if (numValues >= MIN_LENGTH_FOR_ENUM && enumCases.length < Math.sqrt(numValues)) {
            return this.typeBuilder.getEnumType(this.typeName, true, immutable_1.OrderedSet(enumCases));
        }
        return null;
    };
    InferenceUnionBuilder.prototype.makeClass = function (classes, maps) {
        Support_1.assert(maps.length === 0);
        return this._typeInference.inferClassType(this._cjson, this.typeName, this.isInferred, classes);
    };
    InferenceUnionBuilder.prototype.makeArray = function (arrays) {
        return this.typeBuilder.getArrayType(this._typeInference.inferType(this._cjson, pluralize.singular(this.typeName), this.isInferred, arrays));
    };
    return InferenceUnionBuilder;
}(TypeBuilder_1.UnionBuilder));
function canBeEnumCase(s) {
    if (s.length === 0)
        return true; // FIXME: Do we really want this?
    return !DateTime_1.isDate(s) && !DateTime_1.isTime(s, false) && !DateTime_1.isDateTime(s);
}
var TypeInference = /** @class */ (function () {
    function TypeInference(_typeBuilder, _inferEnums) {
        var _this = this;
        this._typeBuilder = _typeBuilder;
        this._inferEnums = _inferEnums;
        this.inferType = function (cjson, typeName, isInferred, valueArray) {
            var unionBuilder = new InferenceUnionBuilder(_this._typeBuilder, typeName, isInferred, _this, cjson);
            var numValues = 0;
            forEachValueInNestedValueArray(valueArray, function (value) {
                numValues += 1;
                var t = CompressedJSON_1.valueTag(value);
                switch (t) {
                    case CompressedJSON_1.Tag.Null:
                        unionBuilder.addNull();
                        break;
                    case CompressedJSON_1.Tag.False:
                    case CompressedJSON_1.Tag.True:
                        unionBuilder.addBool();
                        break;
                    case CompressedJSON_1.Tag.Integer:
                        unionBuilder.addInteger();
                        break;
                    case CompressedJSON_1.Tag.Double:
                        unionBuilder.addDouble();
                        break;
                    case CompressedJSON_1.Tag.InternedString:
                        if (_this._inferEnums && !unionBuilder.haveString) {
                            var s = cjson.getStringForValue(value);
                            if (canBeEnumCase(s)) {
                                unionBuilder.addEnumCase(s);
                            }
                            else {
                                unionBuilder.addStringType("string");
                            }
                        }
                        else {
                            unionBuilder.addStringType("string");
                        }
                        break;
                    case CompressedJSON_1.Tag.UninternedString:
                        unionBuilder.addStringType("string");
                        break;
                    case CompressedJSON_1.Tag.Object:
                        unionBuilder.addClass(cjson.getObjectForValue(value));
                        break;
                    case CompressedJSON_1.Tag.Array:
                        unionBuilder.addArray(cjson.getArrayForValue(value));
                        break;
                    case CompressedJSON_1.Tag.Date:
                        unionBuilder.addStringType("date");
                        break;
                    case CompressedJSON_1.Tag.Time:
                        unionBuilder.addStringType("time");
                        break;
                    case CompressedJSON_1.Tag.DateTime:
                        unionBuilder.addStringType("date-time");
                        break;
                    default:
                        return Support_1.assertNever(t);
                }
            });
            unionBuilder.setNumValues(numValues);
            return unionBuilder.buildUnion(false);
        };
        this.inferClassType = function (cjson, typeName, isInferred, objects) {
            var propertyNames = [];
            var propertyValues = {};
            forEachArrayInNestedValueArray(objects, function (arr) {
                for (var i = 0; i < arr.length; i += 2) {
                    var key = cjson.getStringForValue(arr[i]);
                    var value = arr[i + 1];
                    if (!Object.prototype.hasOwnProperty.call(propertyValues, key)) {
                        propertyNames.push(key);
                        propertyValues[key] = [];
                    }
                    propertyValues[key].push(value);
                }
            });
            var properties = [];
            for (var _i = 0, propertyNames_1 = propertyNames; _i < propertyNames_1.length; _i++) {
                var key = propertyNames_1[_i];
                var values = propertyValues[key];
                var t = _this.inferType(cjson, key, true, values);
                if (values.length < objects.length) {
                    t = _this._typeBuilder.makeNullable(t, key, true);
                }
                properties.push([key, t]);
            }
            var propertyMap = immutable_1.OrderedMap(properties);
            return _this._typeBuilder.getClassType(typeName, isInferred, propertyMap);
        };
    }
    return TypeInference;
}());
exports.TypeInference = TypeInference;
