"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var pluralize = require("pluralize");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var TypeBuilder_1 = require("./TypeBuilder");
var PathElementKind;
(function (PathElementKind) {
    PathElementKind[PathElementKind["Root"] = 0] = "Root";
    PathElementKind[PathElementKind["Definition"] = 1] = "Definition";
    PathElementKind[PathElementKind["OneOf"] = 2] = "OneOf";
    PathElementKind[PathElementKind["AnyOf"] = 3] = "AnyOf";
    PathElementKind[PathElementKind["Property"] = 4] = "Property";
    PathElementKind[PathElementKind["AdditionalProperty"] = 5] = "AdditionalProperty";
    PathElementKind[PathElementKind["Items"] = 6] = "Items";
})(PathElementKind || (PathElementKind = {}));
function checkStringArray(arr) {
    if (!Array.isArray(arr)) {
        return Support_1.panic("Expected a string array, but got " + arr);
    }
    for (var _i = 0, arr_1 = arr; _i < arr_1.length; _i++) {
        var e = arr_1[_i];
        if (typeof e !== "string") {
            return Support_1.panic("Expected string, but got " + e);
        }
    }
    return arr;
}
function parseRef(ref) {
    if (typeof ref !== "string") {
        return Support_1.panic("$ref must be a string");
    }
    var refName = "Something";
    var parts = ref.split("/");
    var elements = [];
    for (var i = 0; i < parts.length; i++) {
        if (parts[i] === "#") {
            elements.push({ kind: PathElementKind.Root });
            refName = "Root";
        }
        else if (parts[i] === "definitions" && i + 1 < parts.length) {
            refName = parts[i + 1];
            elements.push({ kind: PathElementKind.Definition, name: refName });
            i += 1;
        }
        else {
            Support_1.panic("Could not parse JSON schema reference " + ref);
        }
    }
    return [immutable_1.List(elements), refName];
}
function lookupDefinition(schema, name) {
    var definitions = Support_1.checkStringMap(schema.definitions);
    return Support_1.checkStringMap(definitions[name]);
}
function lookupProperty(schema, name) {
    var properties = Support_1.checkStringMap(schema.properties);
    return Support_1.checkStringMap(properties[name]);
}
function indexArray(cases, index) {
    if (!Array.isArray(cases)) {
        return Support_1.panic("oneOf or anyOf value must be an array");
    }
    return Support_1.checkStringMap(cases[index]);
}
function getName(schema, name, isInferred) {
    if (!isInferred) {
        return [name, false];
    }
    var title = schema.title;
    if (typeof title === "string") {
        return [title, false];
    }
    else {
        return [name, true];
    }
}
function checkTypeList(typeOrTypes) {
    if (typeof typeOrTypes === "string") {
        return immutable_1.OrderedSet([typeOrTypes]);
    }
    else if (Array.isArray(typeOrTypes)) {
        var arr = [];
        for (var _i = 0, typeOrTypes_1 = typeOrTypes; _i < typeOrTypes_1.length; _i++) {
            var t = typeOrTypes_1[_i];
            if (typeof t !== "string") {
                return Support_1.panic("element of type is not a string: " + t);
            }
            arr.push(t);
        }
        var set = immutable_1.OrderedSet(arr);
        Support_1.assert(!set.isEmpty(), "JSON Schema must specify at least one type");
        return set;
    }
    else {
        return Support_1.panic("type is neither a string or array of strings: " + typeOrTypes);
    }
}
// Here's a case we can't handle: If the schema specifies
// types
//
//   Foo = class { x: Bar }
//   Bar = Foo | Quux
//   Quux = class { ... }
//
// then to resolve the properties of `Foo` we have to know
// the properties of `Bar`, but to resolve those we have to
// know the properties of `Foo`.
function getHopefullyFinishedType(builder, t) {
    var result = builder.lookupType(t);
    if (result === undefined) {
        return Support_1.panic("Inconveniently recursive types in schema");
    }
    return result;
}
function assertIsClass(t) {
    if (!(t instanceof Type_1.ClassType)) {
        return Support_1.panic("Supposed class type is not a class type");
    }
    return t;
}
function makeImmutablePath(path) {
    return path.map(function (pe) { return immutable_1.fromJS(pe); });
}
var UnifyUnionBuilder = /** @class */ (function (_super) {
    __extends(UnifyUnionBuilder, _super);
    function UnifyUnionBuilder(typeBuilder, typeName, isInferred, _unifyTypes) {
        var _this = _super.call(this, typeBuilder, typeName, isInferred) || this;
        _this._unifyTypes = _unifyTypes;
        return _this;
    }
    UnifyUnionBuilder.prototype.makeEnum = function (enumCases) {
        return this.typeBuilder.getEnumType(this.typeName, this.isInferred, immutable_1.OrderedSet(enumCases));
    };
    UnifyUnionBuilder.prototype.makeClass = function (classes, maps) {
        var _this = this;
        if (classes.length > 0 && maps.length > 0) {
            return Support_1.panic("Cannot handle a class type that's also a map");
        }
        if (maps.length > 0) {
            return this.typeBuilder.getMapType(this._unifyTypes(maps, this.typeName, this.isInferred));
        }
        if (classes.length === 1) {
            return classes[0];
        }
        var properties = immutable_1.OrderedMap();
        var actualClasses = [];
        for (var _i = 0, classes_1 = classes; _i < classes_1.length; _i++) {
            var c = classes_1[_i];
            var t = assertIsClass(getHopefullyFinishedType(this.typeBuilder, c));
            actualClasses.push(t);
        }
        for (var _a = 0, actualClasses_1 = actualClasses; _a < actualClasses_1.length; _a++) {
            var c = actualClasses_1[_a];
            c.properties.forEach(function (t, name) {
                var types = properties.get(name);
                if (types === undefined) {
                    properties = properties.set(name, [t.typeRef]);
                }
                else {
                    types.push(t.typeRef);
                }
            });
        }
        return this.typeBuilder.getUniqueClassType(this.typeName, this.isInferred, properties.map(function (ts, name) { return _this._unifyTypes(ts, name, true); }).sortBy(function (_, n) { return n; }));
    };
    UnifyUnionBuilder.prototype.makeArray = function (arrays) {
        return this.typeBuilder.getArrayType(this._unifyTypes(arrays, pluralize.singular(this.typeName), true));
    };
    return UnifyUnionBuilder;
}(TypeBuilder_1.UnionBuilder));
function schemaToType(typeBuilder, topLevelName, rootJson) {
    var root = Support_1.checkStringMap(rootJson);
    var typeForPath = immutable_1.Map();
    function setTypeForPath(path, t) {
        typeForPath = typeForPath.set(makeImmutablePath(path), t);
    }
    function unifyTypes(typesToUnify, typeName, isInferred) {
        if (typesToUnify.length === 0) {
            return Support_1.panic("Cannot unify empty list of types");
        }
        else if (typesToUnify.length === 1) {
            return typesToUnify[0];
        }
        else {
            var unionBuilder_1 = new UnifyUnionBuilder(typeBuilder, typeName, isInferred, unifyTypes);
            var registerType_1 = function (t) {
                Type_1.matchTypeExhaustive(t, function (_anyType) { return unionBuilder_1.addAny(); }, function (_nullType) { return unionBuilder_1.addNull(); }, function (_boolType) { return unionBuilder_1.addBool(); }, function (_integerType) { return unionBuilder_1.addInteger(); }, function (_doubleType) { return unionBuilder_1.addDouble(); }, function (_stringType) { return unionBuilder_1.addStringType("string"); }, function (arrayType) { return unionBuilder_1.addArray(arrayType.items.typeRef); }, function (classType) { return unionBuilder_1.addClass(classType.typeRef); }, function (mapType) { return unionBuilder_1.addMap(mapType.values.typeRef); }, function (enumType) { return enumType.cases.forEach(function (s) { return unionBuilder_1.addEnumCase(s); }); }, function (unionType) { return unionType.members.forEach(registerType_1); }, function (_dateType) { return unionBuilder_1.addStringType("date"); }, function (_timeType) { return unionBuilder_1.addStringType("time"); }, function (_dateTimeType) { return unionBuilder_1.addStringType("date-time"); });
            };
            for (var _i = 0, typesToUnify_1 = typesToUnify; _i < typesToUnify_1.length; _i++) {
                var t = typesToUnify_1[_i];
                registerType_1(getHopefullyFinishedType(typeBuilder, t));
            }
            return unionBuilder_1.buildUnion(false);
        }
    }
    function lookupRef(local, localPath, ref) {
        var first = ref.first();
        if (first === undefined) {
            return [local, localPath];
        }
        var rest = ref.rest();
        if (first.kind === PathElementKind.Root) {
            return lookupRef(root, immutable_1.List([first]), ref.rest());
        }
        localPath = localPath.push(first);
        switch (first.kind) {
            case PathElementKind.Definition:
                return lookupRef(lookupDefinition(local, first.name), localPath, rest);
            case PathElementKind.OneOf:
                return lookupRef(indexArray(local.oneOf, first.index), localPath, rest);
            case PathElementKind.AnyOf:
                return lookupRef(indexArray(local.anyOf, first.index), localPath, rest);
            case PathElementKind.Property:
                return lookupRef(lookupProperty(local, first.name), localPath, rest);
            case PathElementKind.AdditionalProperty:
                return lookupRef(Support_1.checkStringMap(local.additionalProperties), localPath, rest);
            case PathElementKind.Items:
                return lookupRef(Support_1.checkStringMap(local.items), localPath, rest);
            default:
                return Support_1.assertNever(first);
        }
    }
    function makeClass(schema, path, name, isInferred, properties, requiredArray) {
        var required = immutable_1.Set(requiredArray);
        _a = getName(schema, name, isInferred), name = _a[0], isInferred = _a[1];
        var result = typeBuilder.getUniqueClassType(name, isInferred);
        var c = assertIsClass(getHopefullyFinishedType(typeBuilder, result));
        setTypeForPath(path, result);
        // FIXME: We're using a Map instead of an OrderedMap here because we represent
        // the JSON Schema as a JavaScript object, which has no map ordering.  Ideally
        // we would use a JSON parser that preserves order.
        var props = immutable_1.Map(properties).map(function (propSchema, propName) {
            var t = toType(Support_1.checkStringMap(propSchema), path.push({ kind: PathElementKind.Property, name: propName }), pluralize.singular(propName), true);
            if (!required.has(propName)) {
                return typeBuilder.makeNullable(t, propName, true);
            }
            return t;
        });
        c.setProperties(props.toOrderedMap());
        return result;
        var _a;
    }
    function makeMap(path, name, additional) {
        var valuesType = undefined;
        var mustSet = false;
        var result = typeBuilder.getLazyMapType(function () {
            mustSet = true;
            return valuesType;
        });
        setTypeForPath(path, result);
        path = path.push({ kind: PathElementKind.AdditionalProperty });
        valuesType = toType(additional, path, pluralize.singular(name), true);
        if (mustSet) {
            result.deref().setValues(valuesType);
        }
        return result;
    }
    function fromTypeName(schema, path, inferredName, typeName) {
        var _a = getName(schema, inferredName, true), name = _a[0], isInferred = _a[1];
        switch (typeName) {
            case "object":
                var required = void 0;
                if (schema.required === undefined) {
                    required = [];
                }
                else {
                    required = checkStringArray(schema.required);
                }
                if (schema.properties !== undefined) {
                    return makeClass(schema, path, name, isInferred, Support_1.checkStringMap(schema.properties), required);
                }
                else if (schema.additionalProperties !== undefined) {
                    var additional = schema.additionalProperties;
                    if (additional === true) {
                        return typeBuilder.getMapType(typeBuilder.getPrimitiveType("any"));
                    }
                    else if (additional === false) {
                        return makeClass(schema, path, name, isInferred, {}, required);
                    }
                    else {
                        return makeMap(path, name, Support_1.checkStringMap(additional));
                    }
                }
                else {
                    return typeBuilder.getMapType(typeBuilder.getPrimitiveType("any"));
                }
            case "array":
                if (schema.items !== undefined) {
                    path = path.push({ kind: PathElementKind.Items });
                    return typeBuilder.getArrayType(toType(Support_1.checkStringMap(schema.items), path, pluralize.singular(name), true));
                }
                return typeBuilder.getArrayType(typeBuilder.getPrimitiveType("any"));
            case "boolean":
                return typeBuilder.getPrimitiveType("bool");
            case "string":
                if (schema.format !== undefined) {
                    switch (schema.format) {
                        case "date":
                            return typeBuilder.getPrimitiveType("date");
                        case "time":
                            return typeBuilder.getPrimitiveType("time");
                        case "date-time":
                            return typeBuilder.getPrimitiveType("date-time");
                        default:
                            return Support_1.panic("String format " + schema.format + " not supported");
                    }
                }
                return typeBuilder.getPrimitiveType("string");
            case "null":
                return typeBuilder.getPrimitiveType("null");
            case "integer":
                return typeBuilder.getPrimitiveType("integer");
            case "number":
                return typeBuilder.getPrimitiveType("double");
            default:
                return Support_1.panic("not a type name: " + typeName);
        }
    }
    function convertToType(schema, path, name, isInferred) {
        _a = getName(schema, name, isInferred), name = _a[0], isInferred = _a[1];
        function convertOneOrAnyOf(cases, kind) {
            if (!Array.isArray(cases)) {
                return Support_1.panic("oneOf or anyOf is not an array: " + cases);
            }
            // FIXME: This cast shouldn't be necessary, but TypeScript forces our hand.
            var types = cases.map(function (t, index) {
                return toType(Support_1.checkStringMap(t), path.push({ kind: kind, index: index }), name, isInferred);
            });
            return unifyTypes(types, name, isInferred);
        }
        if (schema.$ref !== undefined) {
            var _b = parseRef(schema.$ref), ref = _b[0], refName = _b[1];
            var _c = lookupRef(schema, path, ref), target = _c[0], targetPath = _c[1];
            return toType(target, targetPath, isInferred ? refName : name, isInferred);
        }
        else if (schema.enum !== undefined) {
            return typeBuilder.getEnumType(name, isInferred, immutable_1.OrderedSet(checkStringArray(schema.enum)));
        }
        else if (schema.type !== undefined) {
            var typeNames = checkTypeList(schema.type);
            if (typeNames.size === 1) {
                return fromTypeName(schema, path, name, Support_1.defined(typeNames.first()));
            }
            else {
                var types = typeNames.map(function (n) { return fromTypeName(schema, path, name, n); });
                return unifyTypes(types.toArray(), name, isInferred);
            }
        }
        else if (schema.oneOf !== undefined) {
            return convertOneOrAnyOf(schema.oneOf, PathElementKind.OneOf);
        }
        else if (schema.anyOf !== undefined) {
            return convertOneOrAnyOf(schema.anyOf, PathElementKind.AnyOf);
        }
        else {
            return typeBuilder.getPrimitiveType("any");
        }
        var _a;
    }
    function toType(schema, path, name, isInferred) {
        // FIXME: This fromJS thing is ugly and inefficient.  Schemas aren't
        // big, so it most likely doesn't matter.
        var immutablePath = makeImmutablePath(path);
        var maybeType = typeForPath.get(immutablePath);
        if (maybeType !== undefined) {
            return maybeType;
        }
        var result = convertToType(schema, path, name, isInferred);
        setTypeForPath(immutablePath, result);
        return result;
    }
    var rootPathElement = { kind: PathElementKind.Root };
    var rootType = toType(root, immutable_1.List([rootPathElement]), topLevelName, false);
    return rootType;
}
exports.schemaToType = schemaToType;
