"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("../Type");
var Source_1 = require("../Source");
var Strings_1 = require("../Strings");
var Support_1 = require("../Support");
var Naming_1 = require("../Naming");
var ConvenienceRenderer_1 = require("../ConvenienceRenderer");
var TargetLanguage_1 = require("../TargetLanguage");
var RendererOptions_1 = require("../RendererOptions");
var Annotation_1 = require("../Annotation");
var JavaTargetLanguage = /** @class */ (function (_super) {
    __extends(JavaTargetLanguage, _super);
    function JavaTargetLanguage() {
        var _this = _super.call(this, "Java", ["java"], "java") || this;
        _this._justTypesOption = new RendererOptions_1.BooleanOption("just-types", "Plain types only", false);
        // FIXME: Do this via a configurable named eventually.
        _this._packageOption = new RendererOptions_1.StringOption("package", "Generated package name", "NAME", "io.quicktype");
        _this.setOptions([_this._packageOption, _this._justTypesOption]);
        return _this;
    }
    Object.defineProperty(JavaTargetLanguage.prototype, "rendererClass", {
        get: function () {
            return JavaRenderer;
        },
        enumerable: true,
        configurable: true
    });
    return JavaTargetLanguage;
}(TargetLanguage_1.TargetLanguage));
exports.default = JavaTargetLanguage;
var keywords = [
    "Object",
    "Class",
    "System",
    "Long",
    "Double",
    "Boolean",
    "String",
    "Map",
    "Exception",
    "IOException",
    "JsonProperty",
    "JsonDeserialize",
    "JsonDeserializer",
    "JsonSerialize",
    "JsonSerializer",
    "JsonParser",
    "JsonProcessingException",
    "DeserializationContext",
    "SerializerProvider",
    "Converter",
    "abstract",
    "continue",
    "for",
    "new",
    "switch",
    "assert",
    "default",
    "goto",
    "package",
    "synchronized",
    "boolean",
    "do",
    "if",
    "private",
    "this",
    "break",
    "double",
    "implements",
    "protected",
    "throw",
    "byte",
    "else",
    "import",
    "public",
    "throws",
    "case",
    "enum",
    "instanceof",
    "return",
    "transient",
    "catch",
    "extends",
    "int",
    "short",
    "try",
    "char",
    "final",
    "interface",
    "static",
    "void",
    "class",
    "finally",
    "long",
    "strictfp",
    "volatile",
    "const",
    "float",
    "native",
    "super",
    "while",
    "null",
    "false",
    "true"
];
var typeNamingFunction = Naming_1.funPrefixNamer(function (n) { return javaNameStyle(true, false, n); });
var propertyNamingFunction = Naming_1.funPrefixNamer(function (n) { return javaNameStyle(false, false, n); });
var enumCaseNamingFunction = Naming_1.funPrefixNamer(function (n) { return javaNameStyle(true, true, n); });
exports.stringEscape = Strings_1.utf16ConcatMap(Strings_1.escapeNonPrintableMapper(Strings_1.isAscii, Strings_1.standardUnicodeHexEscape));
function isStartCharacter(codePoint) {
    if (codePoint === 0x5f)
        return true; // underscore
    return Strings_1.isAscii(codePoint) && Strings_1.isLetter(codePoint);
}
function isPartCharacter(codePoint) {
    return isStartCharacter(codePoint) || (Strings_1.isAscii(codePoint) && Strings_1.isDigit(codePoint));
}
var legalizeName = Strings_1.utf16LegalizeCharacters(isPartCharacter);
// FIXME: Handle acronyms consistently.  In particular, that means that
// we have to use namers to produce the getter and setter names - we can't
// just capitalize and concatenate.
// https://stackoverflow.com/questions/8277355/naming-convention-for-upper-case-abbreviations
function javaNameStyle(startWithUpper, upperUnderscore, original) {
    var words = Strings_1.splitIntoWords(original);
    return Strings_1.combineWords(words, legalizeName, upperUnderscore ? Strings_1.allUpperWordStyle : startWithUpper ? Strings_1.firstUpperWordStyle : Strings_1.allLowerWordStyle, upperUnderscore ? Strings_1.allUpperWordStyle : Strings_1.firstUpperWordStyle, upperUnderscore || startWithUpper ? Strings_1.allUpperWordStyle : Strings_1.allLowerWordStyle, Strings_1.allUpperWordStyle, upperUnderscore ? "_" : "", isStartCharacter);
}
var JavaRenderer = /** @class */ (function (_super) {
    __extends(JavaRenderer, _super);
    function JavaRenderer(graph, _packageName, _justTypes) {
        var _this = _super.call(this, graph) || this;
        _this._packageName = _packageName;
        _this._justTypes = _justTypes;
        _this.fieldOrMethodName = function (methodName, topLevelName) {
            if (_this.topLevels.size === 1) {
                return methodName;
            }
            return [topLevelName, Strings_1.capitalize(methodName)];
        };
        _this.methodName = function (prefix, suffix, topLevelName) {
            if (_this.topLevels.size === 1) {
                return [prefix, suffix];
            }
            return [prefix, topLevelName, suffix];
        };
        _this.decoderName = function (topLevelName) {
            return _this.fieldOrMethodName("fromJsonString", topLevelName);
        };
        _this.encoderName = function (topLevelName) {
            return _this.fieldOrMethodName("toJsonString", topLevelName);
        };
        _this.readerGetterName = function (topLevelName) {
            return _this.methodName("get", "ObjectReader", topLevelName);
        };
        _this.writerGetterName = function (topLevelName) {
            return _this.methodName("get", "ObjectWriter", topLevelName);
        };
        _this.emitFileComment = function (filename) {
            _this.emitLine("// ", filename, ".java");
        };
        _this.emitPackageAndImports = function (imports) {
            var allImports = ["java.util.Map"].concat(_this._justTypes ? [] : imports);
            _this.emitLine("package ", _this._packageName, ";");
            _this.emitNewline();
            for (var _i = 0, allImports_1 = allImports; _i < allImports_1.length; _i++) {
                var pkg = allImports_1[_i];
                _this.emitLine("import ", pkg, ";");
            }
        };
        _this.emitFileHeader = function (fileName, imports) {
            _this.emitFileComment(fileName);
            _this.emitNewline();
            _this.emitPackageAndImports(imports);
            _this.emitNewline();
        };
        _this.emitBlock = function (line, f) {
            _this.emitLine(line, " {");
            _this.indent(f);
            _this.emitLine("}");
        };
        _this.javaType = function (reference, t, withIssues) {
            if (withIssues === void 0) { withIssues = false; }
            return Type_1.matchType(t, function (_anyType) { return Source_1.maybeAnnotated(withIssues, Annotation_1.anyTypeIssueAnnotation, "Object"); }, function (_nullType) { return Source_1.maybeAnnotated(withIssues, Annotation_1.nullTypeIssueAnnotation, "Object"); }, function (_boolType) { return (reference ? "Boolean" : "boolean"); }, function (_integerType) { return (reference ? "Long" : "long"); }, function (_doubleType) { return (reference ? "Double" : "double"); }, function (_stringType) { return "String"; }, function (arrayType) { return [_this.javaType(false, arrayType.items, withIssues), "[]"]; }, function (classType) { return _this.nameForNamedType(classType); }, function (mapType) { return ["Map<String, ", _this.javaType(true, mapType.values, withIssues), ">"]; }, function (enumType) { return _this.nameForNamedType(enumType); }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                if (nullable)
                    return _this.javaType(true, nullable, withIssues);
                return _this.nameForNamedType(unionType);
            });
        };
        _this.javaTypeWithoutGenerics = function (reference, t) {
            if (t instanceof Type_1.ArrayType) {
                return [_this.javaTypeWithoutGenerics(false, t.items), "[]"];
            }
            else if (t instanceof Type_1.MapType) {
                return "Map";
            }
            else if (t instanceof Type_1.UnionType) {
                var nullable = Type_1.nullableFromUnion(t);
                if (nullable)
                    return _this.javaTypeWithoutGenerics(true, nullable);
                return _this.nameForNamedType(t);
            }
            else {
                return _this.javaType(reference, t);
            }
        };
        _this.emitClassDefinition = function (c, className) {
            _this.emitFileHeader(className, ["com.fasterxml.jackson.annotation.*"]);
            if (c.properties.isEmpty() && !_this._justTypes) {
                _this.emitLine("@JsonAutoDetect(fieldVisibility=JsonAutoDetect.Visibility.NONE)");
            }
            _this.emitBlock(["public class ", className], function () {
                _this.forEachClassProperty(c, "none", function (name, _, t) {
                    _this.emitLine("private ", _this.javaType(false, t, true), " ", name, ";");
                });
                _this.forEachClassProperty(c, "leading-and-interposing", function (name, jsonName, t) {
                    if (!_this._justTypes)
                        _this.emitLine('@JsonProperty("', exports.stringEscape(jsonName), '")');
                    var rendered = _this.javaType(false, t);
                    _this.emitLine("public ", rendered, " get", Source_1.modifySource(Strings_1.capitalize, name), "() { return ", name, "; }");
                    if (!_this._justTypes)
                        _this.emitLine('@JsonProperty("', exports.stringEscape(jsonName), '")');
                    _this.emitLine("public void set", Source_1.modifySource(Strings_1.capitalize, name), "(", rendered, " value) { this.", name, " = value; }");
                });
            });
        };
        _this.unionField = function (u, t, withIssues) {
            if (withIssues === void 0) { withIssues = false; }
            var fieldType = _this.javaType(true, t, withIssues);
            // FIXME: "Value" should be part of the name.
            var fieldName = [_this.nameForUnionMember(u, t), "Value"];
            return { fieldType: fieldType, fieldName: fieldName };
        };
        _this.emitUnionDefinition = function (u, unionName) {
            var tokenCase = function (tokenType) {
                _this.emitLine("case ", tokenType, ":");
            };
            var emitNullDeserializer = function () {
                tokenCase("VALUE_NULL");
                _this.indent(function () { return _this.emitLine("break;"); });
            };
            var emitDeserializeType = function (t) {
                var fieldName = _this.unionField(u, t).fieldName;
                var rendered = _this.javaTypeWithoutGenerics(true, t);
                _this.emitLine("value.", fieldName, " = jsonParser.readValueAs(", rendered, ".class);");
                _this.emitLine("break;");
            };
            var emitDeserializer = function (tokenTypes, kind) {
                var t = u.findMember(kind);
                if (!t)
                    return;
                for (var _i = 0, tokenTypes_1 = tokenTypes; _i < tokenTypes_1.length; _i++) {
                    var tokenType = tokenTypes_1[_i];
                    tokenCase(tokenType);
                }
                _this.indent(function () { return emitDeserializeType(t); });
            };
            var emitDoubleSerializer = function () {
                var t = u.findMember("double");
                if (!t)
                    return;
                if (!u.findMember("integer"))
                    tokenCase("VALUE_NUMBER_INT");
                tokenCase("VALUE_NUMBER_FLOAT");
                _this.indent(function () { return emitDeserializeType(t); });
            };
            _this.emitFileHeader(unionName, [
                "java.io.IOException",
                "com.fasterxml.jackson.core.*",
                "com.fasterxml.jackson.databind.*",
                "com.fasterxml.jackson.databind.annotation.*"
            ]);
            if (!_this._justTypes) {
                _this.emitLine("@JsonDeserialize(using = ", unionName, ".Deserializer.class)");
                _this.emitLine("@JsonSerialize(using = ", unionName, ".Serializer.class)");
            }
            var _a = Type_1.removeNullFromUnion(u), maybeNull = _a[0], nonNulls = _a[1];
            _this.emitBlock(["public class ", unionName], function () {
                nonNulls.forEach(function (t) {
                    var _a = _this.unionField(u, t, true), fieldType = _a.fieldType, fieldName = _a.fieldName;
                    _this.emitLine("public ", fieldType, " ", fieldName, ";");
                });
                if (_this._justTypes)
                    return;
                _this.emitNewline();
                _this.emitBlock(["static class Deserializer extends JsonDeserializer<", unionName, ">"], function () {
                    _this.emitLine("@Override");
                    _this.emitBlock([
                        "public ",
                        unionName,
                        " deserialize(JsonParser jsonParser, DeserializationContext deserializationContext) throws IOException, JsonProcessingException"
                    ], function () {
                        _this.emitLine(unionName, " value = new ", unionName, "();");
                        _this.emitLine("switch (jsonParser.getCurrentToken()) {");
                        if (maybeNull)
                            emitNullDeserializer();
                        emitDeserializer(["VALUE_NUMBER_INT"], "integer");
                        emitDoubleSerializer();
                        emitDeserializer(["VALUE_TRUE", "VALUE_FALSE"], "bool");
                        emitDeserializer(["VALUE_STRING"], "string");
                        emitDeserializer(["START_ARRAY"], "array");
                        emitDeserializer(["START_OBJECT"], "class");
                        emitDeserializer(["VALUE_STRING"], "enum");
                        emitDeserializer(["START_OBJECT"], "map");
                        _this.emitLine('default: throw new IOException("Cannot deserialize ', unionName, '");');
                        _this.emitLine("}");
                        _this.emitLine("return value;");
                    });
                });
                _this.emitNewline();
                _this.emitBlock(["static class Serializer extends JsonSerializer<", unionName, ">"], function () {
                    _this.emitLine("@Override");
                    _this.emitBlock([
                        "public void serialize(",
                        unionName,
                        " obj, JsonGenerator jsonGenerator, SerializerProvider serializerProvider) throws IOException"
                    ], function () {
                        nonNulls.forEach(function (t) {
                            var fieldName = _this.unionField(u, t, true).fieldName;
                            _this.emitBlock(["if (obj.", fieldName, " != null)"], function () {
                                _this.emitLine("jsonGenerator.writeObject(obj.", fieldName, ");");
                                _this.emitLine("return;");
                            });
                        });
                        if (maybeNull) {
                            _this.emitLine("jsonGenerator.writeNull();");
                        }
                        else {
                            _this.emitLine('throw new IOException("', unionName, ' must not be null");');
                        }
                    });
                });
            });
        };
        _this.emitEnumDefinition = function (e, enumName) {
            _this.emitFileHeader(enumName, ["java.io.IOException", "com.fasterxml.jackson.annotation.*"]);
            var caseNames = [];
            _this.forEachEnumCase(e, "none", function (name) {
                if (caseNames.length > 0)
                    caseNames.push(", ");
                caseNames.push(name);
            });
            caseNames.push(";");
            _this.emitBlock(["public enum ", enumName], function () {
                _this.emitLine(caseNames);
                _this.emitNewline();
                _this.emitLine("@JsonValue");
                _this.emitBlock("public String toValue()", function () {
                    _this.emitLine("switch (this) {");
                    _this.forEachEnumCase(e, "none", function (name, jsonName) {
                        _this.emitLine("case ", name, ': return "', exports.stringEscape(jsonName), '";');
                    });
                    _this.emitLine("}");
                    _this.emitLine("return null;");
                });
                _this.emitNewline();
                _this.emitLine("@JsonCreator");
                _this.emitBlock(["public static ", enumName, " forValue(String value) throws IOException"], function () {
                    _this.forEachEnumCase(e, "none", function (name, jsonName) {
                        _this.emitLine('if (value.equals("', exports.stringEscape(jsonName), '")) return ', name, ";");
                    });
                    _this.emitLine('throw new IOException("Cannot deserialize ', enumName, '");');
                });
            });
        };
        _this.emitConverterClass = function () {
            _this.emitFileComment("Converter");
            _this.emitNewline();
            _this.emitMultiline("// To use this code, add the following Maven dependency to your project:\n//\n//     com.fasterxml.jackson.core : jackson-databind : 2.9.0\n//\n// Import this package:\n//");
            _this.emitLine("//     import ", _this._packageName, ".Converter;");
            _this.emitMultiline("//\n// Then you can deserialize a JSON string with\n//");
            _this.forEachTopLevel("none", function (t, name) {
                _this.emitLine("//     ", _this.javaType(false, t), " data = Converter.", _this.decoderName(name), "(jsonString);");
            });
            _this.emitNewline();
            _this.emitPackageAndImports([
                "java.io.IOException",
                "com.fasterxml.jackson.databind.*",
                "com.fasterxml.jackson.core.JsonProcessingException"
            ]);
            _this.emitNewline();
            _this.emitBlock(["public class Converter"], function () {
                _this.emitLine("// Serialize/deserialize helpers");
                _this.forEachTopLevel("leading-and-interposing", function (topLevelType, topLevelName) {
                    var topLevelTypeRendered = _this.javaType(false, topLevelType);
                    _this.emitBlock([
                        "public static ",
                        topLevelTypeRendered,
                        " ",
                        _this.decoderName(topLevelName),
                        "(String json) throws IOException"
                    ], function () {
                        _this.emitLine("return ", _this.readerGetterName(topLevelName), "().readValue(json);");
                    });
                    _this.emitNewline();
                    _this.emitBlock([
                        "public static String ",
                        _this.encoderName(topLevelName),
                        "(",
                        topLevelTypeRendered,
                        " obj) throws JsonProcessingException"
                    ], function () {
                        _this.emitLine("return ", _this.writerGetterName(topLevelName), "().writeValueAsString(obj);");
                    });
                });
                _this.forEachTopLevel("leading-and-interposing", function (topLevelType, topLevelName) {
                    var readerName = _this.fieldOrMethodName("reader", topLevelName);
                    var writerName = _this.fieldOrMethodName("writer", topLevelName);
                    _this.emitLine("private static ObjectReader ", readerName, ";");
                    _this.emitLine("private static ObjectWriter ", writerName, ";");
                    _this.emitNewline();
                    _this.emitBlock(["private static void ", _this.methodName("instantiate", "Mapper", topLevelName), "()"], function () {
                        var renderedForClass = _this.javaTypeWithoutGenerics(false, topLevelType);
                        _this.emitLine("ObjectMapper mapper = new ObjectMapper();");
                        _this.emitLine(readerName, " = mapper.reader(", renderedForClass, ".class);");
                        _this.emitLine(writerName, " = mapper.writerFor(", renderedForClass, ".class);");
                    });
                    _this.emitNewline();
                    _this.emitBlock(["private static ObjectReader ", _this.readerGetterName(topLevelName), "()"], function () {
                        _this.emitLine("if (", readerName, " == null) instantiateMapper();");
                        _this.emitLine("return ", readerName, ";");
                    });
                    _this.emitNewline();
                    _this.emitBlock(["private static ObjectWriter ", _this.writerGetterName(topLevelName), "()"], function () {
                        _this.emitLine("if (", writerName, " == null) instantiateMapper();");
                        _this.emitLine("return ", writerName, ";");
                    });
                });
            });
        };
        return _this;
    }
    Object.defineProperty(JavaRenderer.prototype, "forbiddenNamesForGlobalNamespace", {
        get: function () {
            return keywords;
        },
        enumerable: true,
        configurable: true
    });
    JavaRenderer.prototype.forbiddenForClassProperties = function (_c, _classNamed) {
        return { names: [], namespaces: [this.globalNamespace] };
    };
    JavaRenderer.prototype.topLevelNameStyle = function (rawName) {
        return javaNameStyle(true, false, rawName);
    };
    Object.defineProperty(JavaRenderer.prototype, "namedTypeNamer", {
        get: function () {
            return typeNamingFunction;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(JavaRenderer.prototype, "classPropertyNamer", {
        get: function () {
            return propertyNamingFunction;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(JavaRenderer.prototype, "unionMemberNamer", {
        get: function () {
            return propertyNamingFunction;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(JavaRenderer.prototype, "enumCaseNamer", {
        get: function () {
            return enumCaseNamingFunction;
        },
        enumerable: true,
        configurable: true
    });
    JavaRenderer.prototype.unionNeedsName = function (u) {
        return !Type_1.nullableFromUnion(u);
    };
    // FIXME: This is the same as for C#.
    JavaRenderer.prototype.namedTypeToNameForTopLevel = function (type) {
        var definedTypes = type.directlyReachableTypes(function (t) {
            if ((!(t instanceof Type_1.UnionType) && t.isNamedType()) || (t instanceof Type_1.UnionType && !Type_1.nullableFromUnion(t))) {
                return immutable_1.OrderedSet([t]);
            }
            return null;
        });
        Support_1.assert(definedTypes.size <= 1, "Cannot have more than one defined type per top-level");
        // If the top-level type doesn't contain any classes or unions
        // we have to define a class just for the `FromJson` method, in
        // emitFromJsonForTopLevel.
        var first = definedTypes.first();
        if (first === undefined)
            return null;
        return first;
    };
    JavaRenderer.prototype.emitSourceStructure = function () {
        if (!this._justTypes) {
            this.emitConverterClass();
        }
        this.forEachNamedType("leading-and-interposing", false, this.emitClassDefinition, this.emitEnumDefinition, this.emitUnionDefinition);
    };
    return JavaRenderer;
}(ConvenienceRenderer_1.ConvenienceRenderer));
