"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var _ = require("lodash");
var Type_1 = require("../Type");
var Strings_1 = require("../Strings");
var Support_1 = require("../Support");
var Naming_1 = require("../Naming");
var ConvenienceRenderer_1 = require("../ConvenienceRenderer");
var TargetLanguage_1 = require("../TargetLanguage");
var RendererOptions_1 = require("../RendererOptions");
var unicode = require("unicode-properties");
var SimpleTypesTargetLanguage = /** @class */ (function (_super) {
    __extends(SimpleTypesTargetLanguage, _super);
    function SimpleTypesTargetLanguage() {
        var _this = _super.call(this, "Simple Types", ["types"], "txt") || this;
        _this._declareUnionsOption = new RendererOptions_1.BooleanOption("declare-unions", "Declare unions as named types", false);
        _this.setOptions([_this._declareUnionsOption]);
        return _this;
    }
    Object.defineProperty(SimpleTypesTargetLanguage.prototype, "partialStringTypeMapping", {
        get: function () {
            return { date: "date", time: "time", dateTime: "date-time" };
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SimpleTypesTargetLanguage.prototype, "rendererClass", {
        get: function () {
            return SimpleTypesRenderer;
        },
        enumerable: true,
        configurable: true
    });
    return SimpleTypesTargetLanguage;
}(TargetLanguage_1.TargetLanguage));
exports.default = SimpleTypesTargetLanguage;
function isStartCharacter(utf16Unit) {
    return unicode.isAlphabetic(utf16Unit) || utf16Unit === 0x5f; // underscore
}
function isPartCharacter(utf16Unit) {
    var category = unicode.getCategory(utf16Unit);
    return _.includes(["Nd", "Pc", "Mn", "Mc"], category) || isStartCharacter(utf16Unit);
}
var legalizeName = Strings_1.legalizeCharacters(isPartCharacter);
function simpleNameStyle(original, uppercase) {
    var words = Strings_1.splitIntoWords(original);
    return Strings_1.combineWords(words, legalizeName, uppercase ? Strings_1.firstUpperWordStyle : Strings_1.allLowerWordStyle, Strings_1.firstUpperWordStyle, uppercase ? Strings_1.allUpperWordStyle : Strings_1.allLowerWordStyle, Strings_1.allUpperWordStyle, "", isStartCharacter);
}
var SimpleTypesRenderer = /** @class */ (function (_super) {
    __extends(SimpleTypesRenderer, _super);
    function SimpleTypesRenderer(graph, inlineUnions) {
        var _this = _super.call(this, graph) || this;
        _this.inlineUnions = inlineUnions;
        _this.sourceFor = function (t) {
            return Type_1.matchTypeExhaustive(t, function (_anyType) { return "Any"; }, function (_nullType) { return "Null"; }, function (_boolType) { return "Bool"; }, function (_integerType) { return "Int"; }, function (_doubleType) { return "Double"; }, function (_stringType) { return "String"; }, function (arrayType) { return ["List<", _this.sourceFor(arrayType.items), ">"]; }, function (classType) { return _this.nameForNamedType(classType); }, function (mapType) { return ["Map<String, ", _this.sourceFor(mapType.values), ">"]; }, function (enumType) { return _this.nameForNamedType(enumType); }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                if (nullable)
                    return ["Maybe<", _this.sourceFor(nullable), ">"];
                if (_this.inlineUnions) {
                    var children = unionType.children.map(function (c) { return _this.sourceFor(c); });
                    return Support_1.intercalate(" | ", children).toArray();
                }
                else {
                    return _this.nameForNamedType(unionType);
                }
            }, function (_dateType) { return "Date"; }, function (_timeType) { return "Time"; }, function (_dateTimeType) { return "DateTime"; });
        };
        _this.emitClass = function (c, className) {
            _this.emitLine("class ", className, " {");
            _this.indent(function () {
                _this.forEachClassProperty(c, "none", function (name, _jsonName, t) {
                    _this.emitLine(name, ": ", _this.sourceFor(t));
                });
            });
            _this.emitLine("}");
        };
        _this.emitEnum = function (e, enumName) {
            var caseNames = [];
            _this.forEachEnumCase(e, "none", function (name) {
                if (caseNames.length > 0)
                    caseNames.push(" | ");
                caseNames.push(name);
            });
            _this.emitLine("enum ", enumName, " = ", caseNames);
        };
        _this.emitUnion = function (u, unionName) {
            _this.emitLine("union ", unionName, " {");
            _this.indent(function () {
                _this.forEach(u.members, false, false, function (t) {
                    _this.emitLine("case ", _this.sourceFor(t));
                });
            });
            _this.emitLine("}");
        };
        return _this;
    }
    SimpleTypesRenderer.prototype.topLevelNameStyle = function (rawName) {
        return simpleNameStyle(rawName, true);
    };
    Object.defineProperty(SimpleTypesRenderer.prototype, "namedTypeNamer", {
        get: function () {
            return new Naming_1.Namer(function (n) { return simpleNameStyle(n, true); }, []);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SimpleTypesRenderer.prototype, "classPropertyNamer", {
        get: function () {
            return new Naming_1.Namer(function (n) { return simpleNameStyle(n, false); }, []);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SimpleTypesRenderer.prototype, "unionMemberNamer", {
        get: function () {
            return null;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(SimpleTypesRenderer.prototype, "enumCaseNamer", {
        get: function () {
            return new Naming_1.Namer(function (n) { return simpleNameStyle(n, true); }, []);
        },
        enumerable: true,
        configurable: true
    });
    SimpleTypesRenderer.prototype.namedTypeToNameForTopLevel = function (type) {
        if (type.isNamedType()) {
            return type;
        }
        return null;
    };
    SimpleTypesRenderer.prototype.emitSourceStructure = function () {
        this.forEachClass("interposing", this.emitClass);
        this.forEachEnum("leading-and-interposing", this.emitEnum);
        if (!this.inlineUnions) {
            this.forEachUnion("leading-and-interposing", this.emitUnion);
        }
    };
    return SimpleTypesRenderer;
}(ConvenienceRenderer_1.ConvenienceRenderer));
