"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var _ = require("lodash");
var Type_1 = require("../Type");
var Strings_1 = require("../Strings");
var Support_1 = require("../Support");
var Source_1 = require("../Source");
var Naming_1 = require("../Naming");
var ConvenienceRenderer_1 = require("../ConvenienceRenderer");
var TargetLanguage_1 = require("../TargetLanguage");
var RendererOptions_1 = require("../RendererOptions");
var unicode = require("unicode-properties");
var L = /** @class */ (function (_super) {
    __extends(L, _super);
    function L() {
        var _this = _super.call(this, "TypeScript", ["typescript", "ts"], "ts") || this;
        _this._justTypes = new RendererOptions_1.BooleanOption("just-types", "Interfaces only", false);
        _this._declareUnions = new RendererOptions_1.BooleanOption("explicit-unions", "Explicitly name unions", false);
        _this._runtimeTypecheck = new RendererOptions_1.BooleanOption("runtime-typecheck", "Assert JSON.parse results at runtime", false);
        _this.setOptions([_this._justTypes, _this._declareUnions, _this._runtimeTypecheck]);
        return _this;
    }
    Object.defineProperty(L.prototype, "rendererClass", {
        get: function () {
            return TypeScriptRenderer;
        },
        enumerable: true,
        configurable: true
    });
    return L;
}(TargetLanguage_1.TargetLanguage));
exports.default = L;
function isStartCharacter(utf16Unit) {
    return unicode.isAlphabetic(utf16Unit) || utf16Unit === 0x5f; // underscore
}
function isPartCharacter(utf16Unit) {
    var category = unicode.getCategory(utf16Unit);
    return _.includes(["Nd", "Pc", "Mn", "Mc"], category) || isStartCharacter(utf16Unit);
}
var legalizeName = Strings_1.utf16LegalizeCharacters(isPartCharacter);
function typeNameStyle(original) {
    var words = Strings_1.splitIntoWords(original);
    return Strings_1.combineWords(words, legalizeName, Strings_1.firstUpperWordStyle, Strings_1.firstUpperWordStyle, Strings_1.allUpperWordStyle, Strings_1.allUpperWordStyle, "", isStartCharacter);
}
function propertyNameStyle(original) {
    var escaped = Strings_1.stringEscape(original);
    var quoted = "\"" + escaped + "\"";
    if (original.length === 0) {
        return quoted;
    }
    else if (!isStartCharacter(original.codePointAt(0))) {
        return quoted;
    }
    else if (escaped !== original) {
        return quoted;
    }
    else if (legalizeName(original) !== original) {
        return quoted;
    }
    else {
        return original;
    }
}
var TypeScriptRenderer = /** @class */ (function (_super) {
    __extends(TypeScriptRenderer, _super);
    function TypeScriptRenderer(graph, _justTypes, declareUnions, _runtimeTypecheck) {
        var _this = _super.call(this, graph) || this;
        _this._justTypes = _justTypes;
        _this._runtimeTypecheck = _runtimeTypecheck;
        _this.emitEnum = function (e, enumName) {
            _this.emitBlock(["export enum ", enumName], "", function () {
                _this.forEachEnumCase(e, "none", function (name, jsonName) {
                    _this.emitLine(name, " = \"" + Strings_1.stringEscape(jsonName) + "\",");
                });
            });
            _this.emitNewline();
        };
        _this.sourceFor = function (t) {
            return Type_1.matchType(t, function (_anyType) { return "any"; }, function (_nullType) { return "null"; }, function (_boolType) { return "boolean"; }, function (_integerType) { return "number"; }, function (_doubleType) { return "number"; }, function (_stringType) { return "string"; }, function (arrayType) {
                var itemType = _this.sourceFor(arrayType.items);
                if (_this._inlineUnions && arrayType.items instanceof Type_1.UnionType) {
                    var nullable = Type_1.nullableFromUnion(arrayType.items);
                    if (nullable !== null) {
                        return [_this.sourceFor(nullable), "[]"];
                    }
                    else {
                        return ["Array<", itemType, ">"];
                    }
                }
                else if (arrayType.items instanceof Type_1.ArrayType) {
                    return ["Array<", itemType, ">"];
                }
                else {
                    return [itemType, "[]"];
                }
            }, function (classType) { return _this.nameForNamedType(classType); }, function (mapType) { return ["{ [key: string]: ", _this.sourceFor(mapType.values), " }"]; }, function (enumType) { return _this.nameForNamedType(enumType); }, function (unionType) {
                if (_this._inlineUnions || Type_1.nullableFromUnion(unionType)) {
                    var children = unionType.children.map(_this.sourceFor);
                    return Support_1.intercalate(" | ", children).toArray();
                }
                else {
                    return _this.nameForNamedType(unionType);
                }
            });
        };
        _this.typeMapTypeFor = function (t) {
            return Type_1.matchType(t, function (_anyType) { return "undefined"; }, function (_nullType) { return "undefined"; }, function (_boolType) { return "false"; }, function (_integerType) { return "0"; }, function (_doubleType) { return "3.14"; }, function (_stringType) { return "\"\""; }, function (arrayType) { return ["A(", _this.typeMapTypeFor(arrayType.items), ")"]; }, function (classType) { return ['O("', _this.nameForNamedType(classType), '")']; }, function (mapType) { return ["M(", _this.typeMapTypeFor(mapType.values), ")"]; }, function (enumType) { return ['E("', _this.nameForNamedType(enumType), '")']; }, function (unionType) {
                var children = unionType.children.map(_this.typeMapTypeFor);
                return ["U("].concat(Support_1.intercalate(", ", children).toArray(), [")"]);
            });
        };
        _this.emitBlock = function (source, end, emit) {
            _this.emitLine(source, " {");
            _this.indent(emit);
            _this.emitLine("}", end);
        };
        _this.emitTypeMap = function () {
            _this.emitBlock("const typeMap: any =", ";", function () {
                _this.forEachClass("none", function (t, name) {
                    _this.emitBlock(['"', name, '":'], ",", function () {
                        _this.forEachClassProperty(t, "none", function (propName, _propJsonName, propType) {
                            _this.emitLine(propName, ": ", _this.typeMapTypeFor(propType), ",");
                        });
                    });
                });
                _this.forEachEnum("none", function (e, name) {
                    _this.emitLine('"', name, '": [');
                    _this.indent(function () {
                        _this.forEachEnumCase(e, "none", function (caseName) {
                            _this.emitLine(name, ".", caseName, ",");
                        });
                    });
                    _this.emitLine("],");
                });
            });
        };
        _this.emitClass = function (c, className) {
            _this.emitBlock(["export interface ", className], "", function () {
                var table = [];
                _this.forEachClassProperty(c, "none", function (name, _jsonName, t) {
                    var nullable = t instanceof Type_1.UnionType && Type_1.nullableFromUnion(t);
                    table.push([[name, nullable ? "?" : "", ": "], [_this.sourceFor(nullable || t), ";"]]);
                });
                _this.emitTable(table);
            });
            _this.emitNewline();
        };
        _this.emitConvertModule = function () {
            _this.emitMultiline("// Converts JSON strings to/from your types");
            if (_this._runtimeTypecheck) {
                _this.emitMultiline("// and asserts the results of JSON.parse at runtime");
            }
            _this.emitBlock("export module Convert", "", function () {
                if (_this._runtimeTypecheck) {
                    _this.emitLine("let path: string[] = [];");
                    _this.emitNewline();
                }
                _this.forEachTopLevel("interposing", function (t, name) {
                    _this.emitBlock(["export function to", name, "(json: string): ", _this.sourceFor(t)], "", function () {
                        if (_this._runtimeTypecheck) {
                            _this.emitLine("return cast(JSON.parse(json), ", _this.typeMapTypeFor(t), ");");
                        }
                        else {
                            _this.emitLine("return JSON.parse(json);");
                        }
                    });
                    _this.emitNewline();
                    var camelCaseName = Source_1.modifySource(Strings_1.camelCase, name);
                    _this.emitBlock(["export function ", camelCaseName, "ToJson(value: ", _this.sourceFor(t), "): string"], "", function () {
                        _this.emitLine("return JSON.stringify(value, null, 2);");
                    });
                });
                if (_this._runtimeTypecheck) {
                    _this.emitMultiline("\nfunction cast<T>(obj: any, typ: any): T {\n    path = [];\n    if (!isValid(typ, obj)) {\n        throw `Invalid value: obj${path.join(\"\")}`;\n    }\n    return obj;\n}\n\nfunction isValid(typ: any, val: any): boolean {\n    return typ.isUnion  ? isValidUnion(typ.typs, val)\n            : typ.isArray  ? isValidArray(typ.typ, val)\n            : typ.isMap    ? isValidMap(typ.typ, val)\n            : typ.isEnum   ? isValidEnum(typ.name, val)\n            : typ.isObject ? isValidObject(typ.cls, val)\n            :                isValidPrimitive(typ, val);\n}\n\nfunction isValidPrimitive(typ: string, val: any) {\n    return typeof typ === typeof val;\n}\n\nfunction isValidUnion(typs: any[], val: any): boolean {\n    // val must validate against one typ in typs\n    return typs.find(typ => isValid(typ, val)) !== undefined;\n}\n\nfunction isValidEnum(enumName: string, val: any): boolean {\n    const cases = typeMap[enumName];\n    return cases.indexOf(val) !== -1;\n}\n\nfunction isValidArray(typ: any, val: any): boolean {\n    // val must be an array with no invalid elements\n    return Array.isArray(val) && !val.find((element, i) => {\n        path.push(`[${i}]`);\n        if (isValid(typ, element)) {\n            path.pop();\n            return false;\n        } else {\n            return true;\n        }\n    });\n}\n\nfunction isValidMap(typ: any, val: any): boolean {\n    // all values in the map must be typ\n    for (const prop in val) {\n        if (!!prop) continue;\n        path.push(`[\"${prop}\"]`);\n        if (!isValid(typ, val[prop]))\n            return false;\n        path.pop();\n    }\n    return true;\n}\n\nfunction isValidObject(className: string, val: any): boolean {\n    let typeRep = typeMap[className];\n    \n    for (const prop in typeRep) {\n        if (!!prop) continue;\n        path.push(`.${prop}`);\n        if (!isValid(typeRep[prop], val[prop]))\n            return false;\n        path.pop();\n    }\n\n    return true;\n}\n\nfunction A(typ: any) {\n    return { typ, isArray: true };\n}\n\nfunction E(name: string) {\n    return { name, isEnum: true };\n}\n\nfunction U(...typs: any[]) {\n    return { typs, isUnion: true };\n}\n\nfunction M(typ: any) {\n    return { typ, isMap: true };\n}\n\nfunction O(className: string) {\n    return { cls: className, isObject: true };\n}\n");
                    _this.emitTypeMap();
                }
            });
        };
        _this.emitUnion = function (u, unionName) {
            if (_this._inlineUnions) {
                return;
            }
            var children = u.children.map(_this.sourceFor);
            _this.emitLine("export type ", unionName, " = ", Support_1.intercalate(" | ", children).toArray(), ";");
            _this.emitNewline();
        };
        _this._inlineUnions = !declareUnions;
        return _this;
    }
    TypeScriptRenderer.prototype.topLevelNameStyle = function (rawName) {
        return typeNameStyle(rawName);
    };
    Object.defineProperty(TypeScriptRenderer.prototype, "namedTypeNamer", {
        get: function () {
            return new Naming_1.Namer(typeNameStyle, []);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TypeScriptRenderer.prototype, "classPropertyNamer", {
        get: function () {
            return new Naming_1.Namer(propertyNameStyle, []);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TypeScriptRenderer.prototype, "unionMemberNamer", {
        get: function () {
            return null;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TypeScriptRenderer.prototype, "enumCaseNamer", {
        get: function () {
            return new Naming_1.Namer(typeNameStyle, []);
        },
        enumerable: true,
        configurable: true
    });
    TypeScriptRenderer.prototype.namedTypeToNameForTopLevel = function (type) {
        if (type.isNamedType()) {
            return type;
        }
        return null;
    };
    TypeScriptRenderer.prototype.emitSourceStructure = function () {
        var _this = this;
        if (!this._justTypes) {
            this.emitMultiline("// To parse this data:\n//");
            var topLevelNames_1 = [];
            this.forEachTopLevel("none", function (_t, name) {
                topLevelNames_1.push(", ", name);
            }, function (t) { return t.isNamedType(); });
            this.emitLine("//   import { Convert", topLevelNames_1, ' } from "./file";');
            this.emitLine("//");
            this.forEachTopLevel("none", function (_t, name) {
                var camelCaseName = Source_1.modifySource(Strings_1.camelCase, name);
                _this.emitLine("//   const ", camelCaseName, " = Convert.to", name, "(json);");
            });
            if (this._runtimeTypecheck) {
                this.emitLine("//");
                this.emitLine("// These functions will throw an error if the JSON doesn't");
                this.emitLine("// match the expected interface, even if the JSON is valid.");
            }
            this.emitNewline();
        }
        this.forEachNamedType("none", false, this.emitClass, this.emitEnum, this.emitUnion);
        if (!this._justTypes) {
            this.emitConvertModule();
        }
    };
    return TypeScriptRenderer;
}(ConvenienceRenderer_1.ConvenienceRenderer));
