import { Map, Collection } from "immutable";
import { TypeGraph } from "./TypeGraph";
import { Name, Namespace } from "./Naming";
import { Source, Sourcelike } from "./Source";
import { AnnotationData } from "./Annotation";
export declare type RenderResult = {
    rootSource: Source;
    names: Map<Name, string>;
};
export declare type BlankLineLocations = "none" | "interposing" | "leading" | "leading-and-interposing";
export declare abstract class Renderer {
    protected readonly typeGraph: TypeGraph;
    private _names;
    private _lastNewline?;
    private _emitted;
    private _currentEmitTarget;
    constructor(typeGraph: TypeGraph);
    emitNewline(): void;
    emitLine(...lineParts: Sourcelike[]): void;
    emitMultiline(linesString: string): void;
    emitAnnotated(annotation: AnnotationData, emitter: () => void): void;
    emitIssue(message: string, emitter: () => void): void;
    protected emitTable: (tableArray: Sourcelike[][]) => void;
    private changeIndent(offset);
    forEach<K, V>(iterable: Collection<K, V>, interposedBlankLines: boolean, leadingBlankLine: boolean, emitter: (v: V, k: K) => void): void;
    forEachWithBlankLines<K, V>(iterable: Collection<K, V>, blankLineLocations: BlankLineLocations, emitter: (v: V, k: K) => void): void;
    indent(fn: () => void): void;
    finishedSource: () => Source;
    protected abstract setUpNaming(): Namespace[];
    protected abstract emitSource(): void;
    render: () => RenderResult;
    readonly names: Map<Name, string>;
}
