"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Source_1 = require("./Source");
var Support_1 = require("./Support");
var TargetLanguage = /** @class */ (function () {
    function TargetLanguage(displayName, names, extension) {
        var _this = this;
        this.displayName = displayName;
        this.names = names;
        this.extension = extension;
        this.setOptions = function (options) {
            Support_1.assert(_this._options === undefined, "Target language " + _this.displayName + " sets its options more than once");
            _this._options = options;
        };
    }
    Object.defineProperty(TargetLanguage.prototype, "optionDefinitions", {
        get: function () {
            if (this._options === undefined) {
                return Support_1.panic("Target language " + this.displayName + " did not set its options");
            }
            return this._options.map(function (o) { return o.definition; });
        },
        enumerable: true,
        configurable: true
    });
    TargetLanguage.prototype.renderGraphAndSerialize = function (graph, alphabetizeProperties, rendererOptions) {
        if (this._options === undefined) {
            return Support_1.panic("Target language " + this.displayName + " did not set its options");
        }
        var renderer = new ((_a = this.rendererClass).bind.apply(_a, [void 0, graph].concat(this._options.map(function (o) { return o.getValue(rendererOptions); }))))();
        if (renderer.setAlphabetizeProperties !== undefined) {
            renderer.setAlphabetizeProperties(alphabetizeProperties);
        }
        var renderResult = renderer.render();
        return Source_1.serializeRenderResult(renderResult, this.indentation);
        var _a;
    };
    Object.defineProperty(TargetLanguage.prototype, "indentation", {
        get: function () {
            return "    ";
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TargetLanguage.prototype, "partialStringTypeMapping", {
        get: function () {
            return {};
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TargetLanguage.prototype, "stringTypeMapping", {
        get: function () {
            var partial = this.partialStringTypeMapping;
            return {
                date: partial.date || "string",
                time: partial.time || "string",
                dateTime: partial.dateTime || "string"
            };
        },
        enumerable: true,
        configurable: true
    });
    return TargetLanguage;
}());
exports.TargetLanguage = TargetLanguage;
