"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Support_1 = require("./Support");
var Type = /** @class */ (function () {
    function Type(typeRef, kind) {
        this.typeRef = typeRef;
        this.kind = kind;
    }
    Type.prototype.isNamedType = function () {
        return false;
    };
    Object.defineProperty(Type.prototype, "isStringType", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    Type.prototype.directlyReachableTypes = function (setForType) {
        var set = setForType(this);
        if (set)
            return set;
        return orderedSetUnion(this.children.map(function (t) { return t.directlyReachableTypes(setForType); }));
    };
    Type.prototype.equals = function (other) {
        if (!Object.prototype.hasOwnProperty.call(other, "typeRef")) {
            return false;
        }
        return this.typeRef.equals(other.typeRef);
    };
    Type.prototype.hashCode = function () {
        return this.typeRef.hashCode();
    };
    return Type;
}());
exports.Type = Type;
var PrimitiveType = /** @class */ (function (_super) {
    __extends(PrimitiveType, _super);
    function PrimitiveType(typeRef, kind) {
        return _super.call(this, typeRef, kind) || this;
    }
    Object.defineProperty(PrimitiveType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(PrimitiveType.prototype, "isNullable", {
        get: function () {
            return this.kind === "null";
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(PrimitiveType.prototype, "isStringType", {
        get: function () {
            var kind = this.kind;
            return kind === "string" || kind === "date" || kind === "time" || kind === "date-time";
        },
        enumerable: true,
        configurable: true
    });
    PrimitiveType.prototype.map = function (builder, _) {
        return builder.getPrimitiveType(this.kind);
    };
    return PrimitiveType;
}(Type));
exports.PrimitiveType = PrimitiveType;
function isNull(t) {
    return t.kind === "null";
}
var ArrayType = /** @class */ (function (_super) {
    __extends(ArrayType, _super);
    function ArrayType(typeRef, _itemsRef) {
        var _this = _super.call(this, typeRef, "array") || this;
        _this._itemsRef = _itemsRef;
        return _this;
    }
    ArrayType.prototype.setItems = function (itemsRef) {
        if (this._itemsRef !== undefined) {
            return Support_1.panic("Can only set array items once");
        }
        this._itemsRef = itemsRef;
    };
    ArrayType.prototype.getItemsRef = function () {
        if (this._itemsRef === undefined) {
            return Support_1.panic("Array items accessed before they were set");
        }
        return this._itemsRef;
    };
    Object.defineProperty(ArrayType.prototype, "items", {
        get: function () {
            return this.getItemsRef().deref();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ArrayType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet([this.items]);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ArrayType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    ArrayType.prototype.map = function (builder, f) {
        return builder.getArrayType(f(this.getItemsRef()));
    };
    return ArrayType;
}(Type));
exports.ArrayType = ArrayType;
var MapType = /** @class */ (function (_super) {
    __extends(MapType, _super);
    function MapType(typeRef, _valuesRef) {
        var _this = _super.call(this, typeRef, "map") || this;
        _this._valuesRef = _valuesRef;
        return _this;
    }
    MapType.prototype.setValues = function (valuesRef) {
        if (this._valuesRef !== undefined) {
            return Support_1.panic("Can only set map values once");
        }
        this._valuesRef = valuesRef;
    };
    MapType.prototype.getValuesRef = function () {
        if (this._valuesRef === undefined) {
            return Support_1.panic("Map values accessed before they were set");
        }
        return this._valuesRef;
    };
    Object.defineProperty(MapType.prototype, "values", {
        get: function () {
            return this.getValuesRef().deref();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MapType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet([this.values]);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(MapType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    MapType.prototype.map = function (builder, f) {
        return builder.getMapType(f(this.getValuesRef()));
    };
    return MapType;
}(Type));
exports.MapType = MapType;
// FIXME: In the case of overlapping prefixes and suffixes we will
// produce a name that includes the overlap twice.  For example, for
// the names "aaa" and "aaaa" we have the common prefix "aaa" and the
// common suffix "aaa", so we will produce the combined name "aaaaaa".
function combineNames(names) {
    var first = names.first();
    if (first === undefined) {
        return Support_1.panic("Named type has no names");
    }
    if (names.count() === 1) {
        return first;
    }
    var prefixLength = first.length;
    var suffixLength = first.length;
    names.rest().forEach(function (n) {
        prefixLength = Math.min(prefixLength, n.length);
        for (var i = 0; i < prefixLength; i++) {
            if (first[i] !== n[i]) {
                prefixLength = i;
                break;
            }
        }
        suffixLength = Math.min(suffixLength, n.length);
        for (var i = 0; i < suffixLength; i++) {
            if (first[first.length - i - 1] !== n[n.length - i - 1]) {
                suffixLength = i;
                break;
            }
        }
    });
    var prefix = prefixLength > 2 ? first.substr(0, prefixLength) : "";
    var suffix = suffixLength > 2 ? first.substr(first.length - suffixLength) : "";
    var combined = prefix + suffix;
    if (combined.length > 2) {
        return combined;
    }
    return first;
}
function setFromNameOrNames(nameOrNames) {
    if (typeof nameOrNames === "string") {
        return { names: immutable_1.OrderedSet([nameOrNames]), alternatives: immutable_1.OrderedSet() };
    }
    else if (immutable_1.OrderedSet.isOrderedSet(nameOrNames)) {
        return { names: nameOrNames, alternatives: immutable_1.OrderedSet() };
    }
    else {
        return nameOrNames;
    }
}
var NamedType = /** @class */ (function (_super) {
    __extends(NamedType, _super);
    function NamedType(typeRef, kind, nameOrNames, areNamesInferred) {
        var _this = _super.call(this, typeRef, kind) || this;
        var _a = setFromNameOrNames(nameOrNames), names = _a.names, alternatives = _a.alternatives;
        _this._names = names;
        _this._areNamesInferred = areNamesInferred;
        _this._alternativeNames = alternatives;
        return _this;
    }
    NamedType.prototype.isNamedType = function () {
        return true;
    };
    Object.defineProperty(NamedType.prototype, "names", {
        get: function () {
            return this._names;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NamedType.prototype, "areNamesInferred", {
        get: function () {
            return this._areNamesInferred;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NamedType.prototype, "alternativeNames", {
        get: function () {
            return this._alternativeNames;
        },
        enumerable: true,
        configurable: true
    });
    NamedType.prototype.addNames = function (nameOrNames, isInferred) {
        var _a = setFromNameOrNames(nameOrNames), names = _a.names, alternatives = _a.alternatives;
        if (this._areNamesInferred && !isInferred) {
            this._names = names;
            this._areNamesInferred = false;
        }
        else if (this._areNamesInferred === isInferred) {
            this._names = this._names.union(names);
        }
        this._alternativeNames = this._alternativeNames.union(alternatives);
    };
    NamedType.prototype.clearInferredNames = function () {
        if (this._areNamesInferred) {
            this._names = immutable_1.OrderedSet();
        }
        this._alternativeNames = immutable_1.OrderedSet();
    };
    Object.defineProperty(NamedType.prototype, "combinedName", {
        get: function () {
            return combineNames(this._names);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NamedType.prototype, "proposedNames", {
        get: function () {
            return immutable_1.OrderedSet([this.combinedName]).union(this._alternativeNames);
        },
        enumerable: true,
        configurable: true
    });
    return NamedType;
}(Type));
exports.NamedType = NamedType;
var ClassType = /** @class */ (function (_super) {
    __extends(ClassType, _super);
    function ClassType(typeRef, names, areNamesInferred, _propertyRefs) {
        var _this = _super.call(this, typeRef, "class", names, areNamesInferred) || this;
        _this._propertyRefs = _propertyRefs;
        return _this;
    }
    ClassType.prototype.setProperties = function (propertyRefs) {
        if (this._propertyRefs !== undefined) {
            return Support_1.panic("Can only set class properties once");
        }
        this._propertyRefs = propertyRefs;
    };
    ClassType.prototype.getPropertyRefs = function () {
        if (this._propertyRefs === undefined) {
            return Support_1.panic("Class properties accessed before they were set");
        }
        return this._propertyRefs;
    };
    Object.defineProperty(ClassType.prototype, "properties", {
        get: function () {
            return this.getPropertyRefs().map(function (tref) { return tref.deref(); });
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassType.prototype, "sortedProperties", {
        get: function () {
            var properties = this.properties;
            var sortedKeys = properties.keySeq().sort();
            var props = sortedKeys.map(function (k) { return [k, Support_1.defined(properties.get(k))]; });
            return immutable_1.OrderedMap(props);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassType.prototype, "children", {
        get: function () {
            return this.sortedProperties.toOrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ClassType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    ClassType.prototype.map = function (builder, f) {
        var properties = this.getPropertyRefs().map(f);
        return builder.getClassType(this.names, this.areNamesInferred, properties);
    };
    return ClassType;
}(NamedType));
exports.ClassType = ClassType;
var EnumType = /** @class */ (function (_super) {
    __extends(EnumType, _super);
    function EnumType(typeRef, names, areNamesInferred, cases) {
        var _this = _super.call(this, typeRef, "enum", names, areNamesInferred) || this;
        _this.cases = cases;
        return _this;
    }
    Object.defineProperty(EnumType.prototype, "children", {
        get: function () {
            return immutable_1.OrderedSet();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(EnumType.prototype, "isNullable", {
        get: function () {
            return false;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(EnumType.prototype, "isStringType", {
        get: function () {
            return true;
        },
        enumerable: true,
        configurable: true
    });
    EnumType.prototype.map = function (builder, _) {
        return builder.getEnumType(this.names, this.areNamesInferred, this.cases);
    };
    return EnumType;
}(NamedType));
exports.EnumType = EnumType;
var UnionType = /** @class */ (function (_super) {
    __extends(UnionType, _super);
    function UnionType(typeRef, names, areNamesInferred, _memberRefs) {
        var _this = _super.call(this, typeRef, "union", names, areNamesInferred) || this;
        _this._memberRefs = _memberRefs;
        _this.findMember = function (kind) {
            return _this.members.find(function (t) { return t.kind === kind; });
        };
        if (_memberRefs !== undefined) {
            Support_1.assert(_memberRefs.size > 1, "Union has zero members");
        }
        return _this;
    }
    UnionType.prototype.setMembers = function (memberRefs) {
        if (this._memberRefs !== undefined) {
            return Support_1.panic("Can only set map members once");
        }
        Support_1.assert(memberRefs.size > 1, "Union has zero members");
        this._memberRefs = memberRefs;
    };
    UnionType.prototype.getMemberRefs = function () {
        if (this._memberRefs === undefined) {
            return Support_1.panic("Map members accessed before they were set");
        }
        return this._memberRefs;
    };
    Object.defineProperty(UnionType.prototype, "members", {
        get: function () {
            return this.getMemberRefs().map(function (tref) { return tref.deref(); });
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(UnionType.prototype, "stringTypeMembers", {
        get: function () {
            return this.members.filter(function (t) { return t.isStringType; });
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(UnionType.prototype, "children", {
        get: function () {
            return this.sortedMembers;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(UnionType.prototype, "isNullable", {
        get: function () {
            return this.findMember("null") !== undefined;
        },
        enumerable: true,
        configurable: true
    });
    UnionType.prototype.map = function (builder, f) {
        var members = this.getMemberRefs().map(f);
        return builder.getUnionType(this.names, this.areNamesInferred, members);
    };
    Object.defineProperty(UnionType.prototype, "sortedMembers", {
        get: function () {
            // FIXME: We're assuming no two members of the same kind.
            return this.members.sortBy(function (t) { return t.kind; });
        },
        enumerable: true,
        configurable: true
    });
    return UnionType;
}(NamedType));
exports.UnionType = UnionType;
function removeNullFromUnion(t) {
    var nullType = t.findMember("null");
    if (!nullType) {
        return [null, t.members];
    }
    return [nullType, t.members.filterNot(isNull).toOrderedSet()];
}
exports.removeNullFromUnion = removeNullFromUnion;
function nullableFromUnion(t) {
    var _a = removeNullFromUnion(t), hasNull = _a[0], nonNulls = _a[1];
    if (!hasNull)
        return null;
    if (nonNulls.size !== 1)
        return null;
    return Support_1.defined(nonNulls.first());
}
exports.nullableFromUnion = nullableFromUnion;
function nonNullTypeCases(t) {
    if (t.kind === "null") {
        return immutable_1.OrderedSet();
    }
    if (!(t instanceof UnionType)) {
        return immutable_1.OrderedSet([t]);
    }
    var nonNulls = removeNullFromUnion(t)[1];
    return immutable_1.OrderedSet(nonNulls);
}
exports.nonNullTypeCases = nonNullTypeCases;
// FIXME: The outer OrderedSet should be some Collection, but I can't figure out
// which one.  Collection.Indexed doesn't work with OrderedSet, which is unfortunate.
function orderedSetUnion(sets) {
    var setArray = sets.toArray();
    if (setArray.length === 0)
        return immutable_1.OrderedSet();
    if (setArray.length === 1)
        return setArray[0];
    return (_a = setArray[0]).union.apply(_a, setArray.slice(1));
    var _a;
}
function separateNamedTypes(types) {
    var classes = types.filter(function (t) { return t instanceof ClassType; }).toOrderedSet();
    var enums = types.filter(function (t) { return t instanceof EnumType; }).toOrderedSet();
    var unions = types.filter(function (t) { return t instanceof UnionType; }).toOrderedSet();
    return { classes: classes, enums: enums, unions: unions };
}
exports.separateNamedTypes = separateNamedTypes;
function matchTypeExhaustive(t, anyType, nullType, boolType, integerType, doubleType, stringType, arrayType, classType, mapType, enumType, unionType, dateType, timeType, dateTimeType) {
    if (t instanceof PrimitiveType) {
        var f = {
            any: anyType,
            null: nullType,
            bool: boolType,
            integer: integerType,
            double: doubleType,
            string: stringType,
            date: dateType,
            time: timeType,
            "date-time": dateTimeType
        }[t.kind];
        if (f)
            return f(t);
        return Support_1.panic("Unsupported PrimitiveType: " + t.kind);
    }
    else if (t instanceof ArrayType)
        return arrayType(t);
    else if (t instanceof ClassType)
        return classType(t);
    else if (t instanceof MapType)
        return mapType(t);
    else if (t instanceof EnumType)
        return enumType(t);
    else if (t instanceof UnionType)
        return unionType(t);
    return Support_1.panic("Unknown Type");
}
exports.matchTypeExhaustive = matchTypeExhaustive;
function matchType(t, anyType, nullType, boolType, integerType, doubleType, stringType, arrayType, classType, mapType, enumType, unionType, stringTypeMatchers) {
    if (stringTypeMatchers === undefined) {
        stringTypeMatchers = {};
    }
    var typeNotSupported = function (_) {
        return Support_1.panic("Unsupported PrimitiveType");
    };
    return matchTypeExhaustive(t, anyType, nullType, boolType, integerType, doubleType, stringType, arrayType, classType, mapType, enumType, unionType, stringTypeMatchers.dateType || typeNotSupported, stringTypeMatchers.timeType || typeNotSupported, stringTypeMatchers.dateTimeType || typeNotSupported);
}
exports.matchType = matchType;
function matchCompoundType(t, arrayType, classType, mapType, unionType) {
    function ignore(_) {
        return;
    }
    return matchType(t, ignore, ignore, ignore, ignore, ignore, ignore, arrayType, classType, mapType, ignore, unionType, { dateType: ignore, timeType: ignore, dateTimeType: ignore });
}
exports.matchCompoundType = matchCompoundType;
