import { Map, OrderedMap, OrderedSet, List, Set } from "immutable";
import { PrimitiveTypeKind, Type, NameOrNames, PrimitiveStringTypeKind } from "./Type";
import { TypeGraph } from "./TypeGraph";
export declare type TypeRefCallback = (index: number) => void;
export declare class TypeRef {
    readonly graph: TypeGraph;
    private _maybeIndexOrRef?;
    private _callbacks?;
    constructor(graph: TypeGraph, index?: number);
    private follow();
    readonly maybeIndex: number | undefined;
    readonly index: number;
    callWhenResolved: (callback: TypeRefCallback) => void;
    resolve: (tref: TypeRef) => void;
    deref: () => Type;
    equals: (other: any) => boolean;
    hashCode: () => number;
}
export declare type StringTypeMapping = {
    date: PrimitiveStringTypeKind;
    time: PrimitiveStringTypeKind;
    dateTime: PrimitiveStringTypeKind;
};
export declare abstract class TypeBuilder {
    private readonly _stringTypeMapping;
    readonly typeGraph: TypeGraph;
    protected topLevels: Map<string, TypeRef>;
    protected types: List<Type | undefined>;
    protected namesToAdd: Map<number, {
        names: NameOrNames;
        isInferred: boolean;
    }[]>;
    constructor(_stringTypeMapping: StringTypeMapping);
    addTopLevel: (name: string, tref: TypeRef) => void;
    private reserveTypeRef;
    private commitType;
    protected addType<T extends Type>(creator: (tref: TypeRef) => T): TypeRef;
    typeAtIndex: (index: number) => Type;
    addNames: (tref: TypeRef, names: NameOrNames, isInferred: boolean) => void;
    makeNullable: (tref: TypeRef, typeNames: NameOrNames, areNamesInferred: boolean) => TypeRef;
    finish(): TypeGraph;
    private _primitiveTypes;
    private _mapTypes;
    private _arrayTypes;
    private _enumTypes;
    private _classTypes;
    private _unionTypes;
    getPrimitiveType(kind: PrimitiveTypeKind): TypeRef;
    getEnumType(names: NameOrNames, isInferred: boolean, cases: OrderedSet<string>): TypeRef;
    getMapType(values: TypeRef): TypeRef;
    getArrayType(items: TypeRef): TypeRef;
    getClassType(names: NameOrNames, isInferred: boolean, properties: OrderedMap<string, TypeRef>): TypeRef;
    getUnionType(names: NameOrNames, isInferred: boolean, members: OrderedSet<TypeRef>): TypeRef;
}
export declare class TypeGraphBuilder extends TypeBuilder {
    protected typeForEntry(entry: Type | undefined): Type | undefined;
    getLazyMapType(valuesCreator: () => TypeRef | undefined): TypeRef;
    getUniqueClassType: (names: NameOrNames, isInferred: boolean, properties?: OrderedMap<string, TypeRef> | undefined) => TypeRef;
    getUniqueUnionType: (name: string, isInferred: boolean, members: OrderedSet<TypeRef>) => TypeRef;
    lookupType: (typeRef: TypeRef) => Type | undefined;
}
export declare class GraphRewriteBuilder<T extends Type> extends TypeBuilder {
    private readonly _originalGraph;
    private readonly _replacer;
    private _setsToReplaceByMember;
    private _reconstitutedTypes;
    constructor(_originalGraph: TypeGraph, stringTypeMapping: StringTypeMapping, setsToReplace: T[][], _replacer: (typesToReplace: Set<T>, builder: GraphRewriteBuilder<T>) => TypeRef);
    followIndex(index: number): number;
    protected typeForEntry(entry: Type | undefined | number): Type | undefined;
    private withForwardingRef(typeCreator);
    private replaceSet(typesToReplace);
    private getReconstitutedType;
    reconstituteType: (t: Type) => TypeRef;
    finish(): TypeGraph;
}
export declare abstract class UnionBuilder<TArray, TClass, TMap> {
    protected readonly typeBuilder: TypeGraphBuilder;
    protected readonly typeName: string;
    protected readonly isInferred: boolean;
    private _haveAny;
    private _haveNull;
    private _haveBool;
    private _haveInteger;
    private _haveDouble;
    private _stringTypes;
    private readonly _arrays;
    private readonly _maps;
    private readonly _classes;
    private _enumCaseMap;
    private _enumCases;
    constructor(typeBuilder: TypeGraphBuilder, typeName: string, isInferred: boolean);
    readonly haveString: boolean;
    addAny: () => void;
    addNull: () => void;
    addBool: () => void;
    addInteger: () => void;
    addDouble: () => void;
    addStringType: (kind: PrimitiveStringTypeKind) => void;
    addArray: (t: TArray) => void;
    addClass: (t: TClass) => void;
    addMap: (t: TMap) => void;
    addEnumCase: (s: string) => void;
    protected abstract makeEnum(cases: string[]): TypeRef | null;
    protected abstract makeClass(classes: TClass[], maps: TMap[]): TypeRef;
    protected abstract makeArray(arrays: TArray[]): TypeRef;
    buildUnion: (unique: boolean) => TypeRef;
}
