"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var TypeGraph_1 = require("./TypeGraph");
var Support_1 = require("./Support");
var TypeRef = /** @class */ (function () {
    // FIXME: This should refer to the TypeGraph, not the builder.
    // Maybe before the TypeGraph is frozen is holds a reference to
    // its TypeBuilder so it can get the types from there?
    function TypeRef(graph, index) {
        var _this = this;
        this.graph = graph;
        this.callWhenResolved = function (callback) {
            if (_this._maybeIndexOrRef === undefined) {
                if (_this._callbacks === undefined) {
                    _this._callbacks = [];
                }
                _this._callbacks.push(callback);
            }
            else if (typeof _this._maybeIndexOrRef === "number") {
                callback(_this._maybeIndexOrRef);
            }
            else {
                _this._maybeIndexOrRef.callWhenResolved(callback);
            }
        };
        this.resolve = function (tref) {
            if (_this._maybeIndexOrRef !== undefined) {
                return Support_1.panic("Trying to resolve an already resolved type reference");
            }
            Support_1.assert(tref.follow() !== _this, "Tried to create a TypeRef cycle");
            _this._maybeIndexOrRef = tref;
            if (_this._callbacks !== undefined) {
                for (var _i = 0, _a = _this._callbacks; _i < _a.length; _i++) {
                    var cb = _a[_i];
                    tref.callWhenResolved(cb);
                }
                _this._callbacks = undefined;
            }
        };
        this.deref = function () {
            return _this.graph.typeAtIndex(_this.index);
        };
        this.equals = function (other) {
            if (!(other instanceof TypeRef)) {
                return false;
            }
            Support_1.assert(_this.graph === other.graph, "Comparing type refs of different graphs");
            return _this.follow() === other.follow();
        };
        this.hashCode = function () {
            return _this.index | 0;
        };
        this._maybeIndexOrRef = index;
    }
    TypeRef.prototype.follow = function () {
        if (this._maybeIndexOrRef instanceof TypeRef) {
            return this._maybeIndexOrRef.follow();
        }
        return this;
    };
    Object.defineProperty(TypeRef.prototype, "maybeIndex", {
        get: function () {
            var tref = this.follow();
            if (typeof tref._maybeIndexOrRef === "number") {
                return tref._maybeIndexOrRef;
            }
            return undefined;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TypeRef.prototype, "index", {
        get: function () {
            var maybeIndex = this.maybeIndex;
            if (maybeIndex === undefined) {
                return Support_1.panic("Trying to dereference unresolved type reference");
            }
            return maybeIndex;
        },
        enumerable: true,
        configurable: true
    });
    return TypeRef;
}());
exports.TypeRef = TypeRef;
var TypeBuilder = /** @class */ (function () {
    function TypeBuilder(_stringTypeMapping) {
        var _this = this;
        this._stringTypeMapping = _stringTypeMapping;
        this.typeGraph = new TypeGraph_1.TypeGraph(this);
        this.topLevels = immutable_1.Map();
        this.types = immutable_1.List();
        this.namesToAdd = immutable_1.Map();
        this.addTopLevel = function (name, tref) {
            // assert(t.typeGraph === this.typeGraph, "Adding top-level to wrong type graph");
            Support_1.assert(!_this.topLevels.has(name), "Trying to add top-level with existing name");
            Support_1.assert(_this.types.get(tref.index) !== undefined, "Trying to add a top-level type that doesn't exist (yet?)");
            _this.topLevels = _this.topLevels.set(name, tref);
        };
        this.reserveTypeRef = function () {
            var index = _this.types.size;
            _this.types = _this.types.push(undefined);
            return new TypeRef(_this.typeGraph, index);
        };
        this.commitType = function (tref, t) {
            Support_1.assert(_this.types.get(tref.index) === undefined, "A type index was committed twice");
            _this.types = _this.types.set(tref.index, t);
        };
        this.typeAtIndex = function (index) {
            var maybeType = _this.types.get(index);
            if (maybeType === undefined) {
                return Support_1.panic("Trying to deref an undefined type in a type builder");
            }
            return maybeType;
        };
        this.addNames = function (tref, names, isInferred) {
            tref.callWhenResolved(function (index) {
                var t = _this.types.get(index);
                if (t !== undefined) {
                    if (!t.isNamedType()) {
                        return;
                    }
                    t.addNames(names, isInferred);
                }
                else {
                    var entries = _this.namesToAdd.get(index);
                    if (entries === undefined) {
                        entries = [];
                        _this.namesToAdd = _this.namesToAdd.set(index, entries);
                    }
                    entries.push({ names: names, isInferred: isInferred });
                }
            });
        };
        this.makeNullable = function (tref, typeNames, areNamesInferred) {
            var t = Support_1.defined(_this.types.get(tref.index));
            if (t.kind === "null") {
                return tref;
            }
            var nullType = _this.getPrimitiveType("null");
            if (!(t instanceof Type_1.UnionType)) {
                return _this.getUnionType(typeNames, areNamesInferred, immutable_1.OrderedSet([tref, nullType]));
            }
            var _a = Type_1.removeNullFromUnion(t), maybeNull = _a[0], nonNulls = _a[1];
            if (maybeNull)
                return tref;
            return _this.getUnionType(typeNames, areNamesInferred, nonNulls.map(function (nn) { return nn.typeRef; }).add(nullType));
        };
        // FIXME: make mutable?
        this._primitiveTypes = immutable_1.Map();
        this._mapTypes = immutable_1.Map();
        this._arrayTypes = immutable_1.Map();
        this._enumTypes = immutable_1.Map();
        this._classTypes = immutable_1.Map();
        this._unionTypes = immutable_1.Map();
    }
    TypeBuilder.prototype.addType = function (creator) {
        var tref = this.reserveTypeRef();
        var t = creator(tref);
        this.commitType(tref, t);
        var namesToAdd = this.namesToAdd.get(tref.index);
        if (namesToAdd !== undefined) {
            if (t.isNamedType()) {
                for (var _i = 0, namesToAdd_1 = namesToAdd; _i < namesToAdd_1.length; _i++) {
                    var nta = namesToAdd_1[_i];
                    t.addNames(nta.names, nta.isInferred);
                }
            }
            this.namesToAdd = this.namesToAdd.remove(tref.index);
        }
        return tref;
    };
    TypeBuilder.prototype.finish = function () {
        Support_1.assert(this.namesToAdd.isEmpty(), "We're finishing, but still names to add left");
        this.typeGraph.freeze(this.topLevels, this.types.map(Support_1.defined));
        return this.typeGraph;
    };
    TypeBuilder.prototype.getPrimitiveType = function (kind) {
        if (kind === "date")
            kind = this._stringTypeMapping.date;
        if (kind === "time")
            kind = this._stringTypeMapping.time;
        if (kind === "date-time")
            kind = this._stringTypeMapping.dateTime;
        var tref = this._primitiveTypes.get(kind);
        if (tref === undefined) {
            tref = this.addType(function (tr) { return new Type_1.PrimitiveType(tr, kind); });
            this._primitiveTypes = this._primitiveTypes.set(kind, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.getEnumType = function (names, isInferred, cases) {
        var unorderedCases = cases.toSet();
        var tref = this._enumTypes.get(unorderedCases);
        if (tref === undefined) {
            tref = this.addType(function (tr) { return new Type_1.EnumType(tr, names, isInferred, cases); });
            this._enumTypes = this._enumTypes.set(unorderedCases, tref);
        }
        else {
            this.addNames(tref, names, isInferred);
        }
        return tref;
    };
    TypeBuilder.prototype.getMapType = function (values) {
        var tref = this._mapTypes.get(values);
        if (tref === undefined) {
            tref = this.addType(function (tr) { return new Type_1.MapType(tr, values); });
            this._mapTypes = this._mapTypes.set(values, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.getArrayType = function (items) {
        var tref = this._arrayTypes.get(items);
        if (tref === undefined) {
            tref = this.addType(function (tr) { return new Type_1.ArrayType(tr, items); });
            this._arrayTypes = this._arrayTypes.set(items, tref);
        }
        return tref;
    };
    TypeBuilder.prototype.getClassType = function (names, isInferred, properties) {
        var tref = this._classTypes.get(properties.toMap());
        if (tref === undefined) {
            tref = this.addType(function (tr) { return new Type_1.ClassType(tr, names, isInferred, properties); });
            this._classTypes = this._classTypes.set(properties.toMap(), tref);
        }
        else {
            this.addNames(tref, names, isInferred);
        }
        return tref;
    };
    TypeBuilder.prototype.getUnionType = function (names, isInferred, members) {
        var unorderedMembers = members.toSet();
        var tref = this._unionTypes.get(unorderedMembers);
        if (tref === undefined) {
            tref = this.addType(function (tr) { return new Type_1.UnionType(tr, names, isInferred, members); });
            this._unionTypes = this._unionTypes.set(unorderedMembers, tref);
        }
        else {
            this.addNames(tref, names, isInferred);
        }
        return tref;
    };
    return TypeBuilder;
}());
exports.TypeBuilder = TypeBuilder;
var TypeGraphBuilder = /** @class */ (function (_super) {
    __extends(TypeGraphBuilder, _super);
    function TypeGraphBuilder() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.getUniqueClassType = function (names, isInferred, properties) {
            return _this.addType(function (tref) { return new Type_1.ClassType(tref, names, isInferred, properties); });
        };
        _this.getUniqueUnionType = function (name, isInferred, members) {
            return _this.addType(function (tref) { return new Type_1.UnionType(tref, name, isInferred, members); });
        };
        _this.lookupType = function (typeRef) {
            var maybeIndex = typeRef.maybeIndex;
            if (maybeIndex === undefined) {
                return undefined;
            }
            return _this.types.get(maybeIndex);
        };
        return _this;
    }
    TypeGraphBuilder.prototype.typeForEntry = function (entry) {
        return entry;
    };
    TypeGraphBuilder.prototype.getLazyMapType = function (valuesCreator) {
        return this.addType(function (tref) { return new Type_1.MapType(tref, valuesCreator()); });
    };
    return TypeGraphBuilder;
}(TypeBuilder));
exports.TypeGraphBuilder = TypeGraphBuilder;
var GraphRewriteBuilder = /** @class */ (function (_super) {
    __extends(GraphRewriteBuilder, _super);
    function GraphRewriteBuilder(_originalGraph, stringTypeMapping, setsToReplace, _replacer) {
        var _this = _super.call(this, stringTypeMapping) || this;
        _this._originalGraph = _originalGraph;
        _this._replacer = _replacer;
        _this._reconstitutedTypes = immutable_1.Map();
        _this.getReconstitutedType = function (originalRef) {
            var maybeTypeRef = _this._reconstitutedTypes.get(originalRef.index);
            if (maybeTypeRef !== undefined) {
                return maybeTypeRef;
            }
            var maybeSet = _this._setsToReplaceByMember.get(originalRef.index);
            if (maybeSet !== undefined) {
                return _this.replaceSet(maybeSet);
            }
            return _this.withForwardingRef(function (forwardingRef) {
                _this._reconstitutedTypes = _this._reconstitutedTypes.set(originalRef.index, forwardingRef);
                return originalRef.deref().map(_this, _this.getReconstitutedType);
            });
        };
        _this.reconstituteType = function (t) {
            Support_1.assert(t.typeRef.graph === _this._originalGraph, "Trying to reconstitute a type from the wrong graph");
            return _this.getReconstitutedType(t.typeRef);
        };
        _this._setsToReplaceByMember = immutable_1.Map();
        var _loop_1 = function (types) {
            var set = immutable_1.Set(types);
            set.forEach(function (t) {
                var index = t.typeRef.index;
                Support_1.assert(!_this._setsToReplaceByMember.has(index), "A type is member of more than one set to be replaced");
                _this._setsToReplaceByMember = _this._setsToReplaceByMember.set(index, set);
            });
        };
        for (var _i = 0, setsToReplace_1 = setsToReplace; _i < setsToReplace_1.length; _i++) {
            var types = setsToReplace_1[_i];
            _loop_1(types);
        }
        return _this;
    }
    GraphRewriteBuilder.prototype.followIndex = function (index) {
        var entry = this.types.get(index);
        if (typeof entry === "number") {
            return this.followIndex(entry);
        }
        return index;
    };
    GraphRewriteBuilder.prototype.typeForEntry = function (entry) {
        if (typeof entry === "number") {
            entry = this.types.get(this.followIndex(entry));
            if (typeof entry === "number") {
                return Support_1.panic("followIndex led us to a forwarding entry");
            }
        }
        return entry;
    };
    GraphRewriteBuilder.prototype.withForwardingRef = function (typeCreator) {
        var forwardingRef = new TypeRef(this.typeGraph);
        var actualRef = typeCreator(forwardingRef);
        forwardingRef.resolve(actualRef);
        return actualRef;
    };
    GraphRewriteBuilder.prototype.replaceSet = function (typesToReplace) {
        var _this = this;
        return this.withForwardingRef(function (forwardingRef) {
            typesToReplace.forEach(function (t) {
                var originalRef = t.typeRef;
                _this._reconstitutedTypes = _this._reconstitutedTypes.set(originalRef.index, forwardingRef);
                _this._setsToReplaceByMember = _this._setsToReplaceByMember.remove(originalRef.index);
            });
            return _this._replacer(typesToReplace, _this);
        });
    };
    GraphRewriteBuilder.prototype.finish = function () {
        var _this = this;
        this._originalGraph.topLevels.forEach(function (t, name) {
            _this.addTopLevel(name, _this.getReconstitutedType(t.typeRef));
        });
        return _super.prototype.finish.call(this);
    };
    return GraphRewriteBuilder;
}(TypeBuilder));
exports.GraphRewriteBuilder = GraphRewriteBuilder;
var UnionBuilder = /** @class */ (function () {
    function UnionBuilder(typeBuilder, typeName, isInferred) {
        var _this = this;
        this.typeBuilder = typeBuilder;
        this.typeName = typeName;
        this.isInferred = isInferred;
        this._haveAny = false;
        this._haveNull = false;
        this._haveBool = false;
        this._haveInteger = false;
        this._haveDouble = false;
        this._stringTypes = immutable_1.OrderedSet();
        this._arrays = [];
        this._maps = [];
        this._classes = [];
        this._enumCaseMap = {};
        this._enumCases = [];
        this.addAny = function () {
            _this._haveAny = true;
        };
        this.addNull = function () {
            _this._haveNull = true;
        };
        this.addBool = function () {
            _this._haveBool = true;
        };
        this.addInteger = function () {
            _this._haveInteger = true;
        };
        this.addDouble = function () {
            _this._haveDouble = true;
        };
        this.addStringType = function (kind) {
            if (_this._stringTypes.has(kind))
                return;
            // string overrides all other string types, as well as enum
            if (kind === "string") {
                _this._stringTypes = immutable_1.OrderedSet([kind]);
                _this._enumCaseMap = {};
                _this._enumCases = [];
            }
            else {
                if (_this.haveString)
                    return;
                _this._stringTypes = _this._stringTypes.add(kind);
            }
        };
        this.addArray = function (t) {
            _this._arrays.push(t);
        };
        this.addClass = function (t) {
            _this._classes.push(t);
        };
        this.addMap = function (t) {
            _this._maps.push(t);
        };
        this.addEnumCase = function (s) {
            if (_this.haveString) {
                return;
            }
            if (!Object.prototype.hasOwnProperty.call(_this._enumCaseMap, s)) {
                _this._enumCaseMap[s] = 0;
                _this._enumCases.push(s);
            }
            _this._enumCaseMap[s] += 1;
        };
        this.buildUnion = function (unique) {
            var types = [];
            if (_this._haveAny) {
                return _this.typeBuilder.getPrimitiveType("any");
            }
            if (_this._haveNull) {
                types.push(_this.typeBuilder.getPrimitiveType("null"));
            }
            if (_this._haveBool) {
                types.push(_this.typeBuilder.getPrimitiveType("bool"));
            }
            if (_this._haveDouble) {
                types.push(_this.typeBuilder.getPrimitiveType("double"));
            }
            else if (_this._haveInteger) {
                types.push(_this.typeBuilder.getPrimitiveType("integer"));
            }
            _this._stringTypes.forEach(function (kind) {
                types.push(_this.typeBuilder.getPrimitiveType(kind));
            });
            if (_this._enumCases.length > 0) {
                var maybeEnum = _this.makeEnum(_this._enumCases);
                if (maybeEnum !== null) {
                    types.push(maybeEnum);
                }
                else {
                    types.push(_this.typeBuilder.getPrimitiveType("string"));
                }
            }
            if (_this._classes.length > 0 || _this._maps.length > 0) {
                types.push(_this.makeClass(_this._classes, _this._maps));
            }
            if (_this._arrays.length > 0) {
                types.push(_this.makeArray(_this._arrays));
            }
            if (types.length === 0) {
                return _this.typeBuilder.getPrimitiveType("any");
            }
            if (types.length === 1) {
                _this.typeBuilder.addNames(types[0], _this.typeName, _this.isInferred);
                return types[0];
            }
            var typesSet = immutable_1.OrderedSet(types);
            if (unique) {
                return _this.typeBuilder.getUniqueUnionType(_this.typeName, _this.isInferred, typesSet);
            }
            else {
                return _this.typeBuilder.getUnionType(_this.typeName, _this.isInferred, typesSet);
            }
        };
    }
    Object.defineProperty(UnionBuilder.prototype, "haveString", {
        get: function () {
            return this._stringTypes.has("string");
        },
        enumerable: true,
        configurable: true
    });
    return UnionBuilder;
}());
exports.UnionBuilder = UnionBuilder;
