/// <reference types="node" />
import { Readable } from "stream";
import { TargetLanguage } from "./TargetLanguage";
import { SerializedRenderResult } from "./Source";
export { TargetLanguage } from "./TargetLanguage";
export { SerializedRenderResult, Annotation } from "./Source";
export { all as languages, languageNamed } from "./Language/All";
export { OptionDefinition } from "./RendererOptions";
export declare function getTargetLanguage(name: string): TargetLanguage;
export declare type RendererOptions = {
    [name: string]: string;
};
export interface Sample<T> {
    source: T;
}
export interface Source<T> {
    name: string;
    samples: Sample<T>[];
}
export interface SchemaData<T> {
    name: string;
    schema: T;
}
export interface GraphQLData {
    name: string;
    schema: any;
    query: string;
}
export declare type SourceType<T> = GraphQLData[] | Source<T>[] | SchemaData<T>[];
export interface Options {
    lang: string;
    sources: SourceType<string | Readable>;
    inferMaps: boolean;
    inferEnums: boolean;
    alphabetizeProperties: boolean;
    combineClasses: boolean;
    noRender: boolean;
    rendererOptions: RendererOptions;
}
export declare class Run {
    private readonly _doCache;
    private _compressedJSON;
    private _allInputs;
    private _options;
    constructor(options: Partial<Options>, _doCache: boolean);
    private readonly isInputJSONSchema;
    private readonly isInputGraphQL;
    private makeGraph;
    private readSampleFromStream;
    run: () => Promise<SerializedRenderResult>;
}
export declare function quicktype(options: Partial<Options>, useCache?: boolean): Promise<SerializedRenderResult>;
