"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var get_stream_1 = require("./get-stream");
var _ = require("lodash");
var immutable_1 = require("immutable");
var targetLanguages = require("./Language/All");
var Support_1 = require("./Support");
var CompressedJSON_1 = require("./CompressedJSON");
var CombineClasses_1 = require("./CombineClasses");
var JSONSchemaInput_1 = require("./JSONSchemaInput");
var Inference_1 = require("./Inference");
var InferMaps_1 = require("./InferMaps");
var TypeBuilder_1 = require("./TypeBuilder");
var GraphQL_1 = require("./GraphQL");
var GatherNames_1 = require("./GatherNames");
// Re-export essential types and functions
var TargetLanguage_1 = require("./TargetLanguage");
exports.TargetLanguage = TargetLanguage_1.TargetLanguage;
var All_1 = require("./Language/All");
exports.languages = All_1.all;
exports.languageNamed = All_1.languageNamed;
var stringToStream = require("string-to-stream");
function getTargetLanguage(name) {
    var language = targetLanguages.languageNamed(name);
    if (language) {
        return language;
    }
    throw new Error("'" + name + "' is not yet supported as an output language.");
}
exports.getTargetLanguage = getTargetLanguage;
function isSourceData(sources) {
    if (_.isArray(sources)) {
        if (sources.length === 0) {
            Support_1.panic("You must provide at least one sample");
        }
        return "samples" in sources[0];
    }
    return false;
}
function isSchemaData(sources) {
    if (_.isArray(sources)) {
        if (sources.length === 0) {
            Support_1.panic("You must provide at least one sample");
        }
        return !("query" in sources[0]) && !("samples" in sources[0]);
    }
    return false;
}
function isGraphQLData(sources) {
    if (_.isArray(sources)) {
        if (sources.length === 0) {
            Support_1.panic("You must provide at least one sample");
        }
        return "query" in sources[0];
    }
    return false;
}
var defaultOptions = {
    lang: "ts",
    sources: [],
    inferMaps: true,
    inferEnums: true,
    alphabetizeProperties: false,
    combineClasses: true,
    noRender: false,
    rendererOptions: {}
};
var graphByInputHash = immutable_1.Map();
function toReadable(source) {
    return _.isString(source) ? stringToStream(source) : source;
}
var Run = /** @class */ (function () {
    function Run(options, _doCache) {
        var _this = this;
        this._doCache = _doCache;
        this.makeGraph = function () {
            var stringTypeMapping = getTargetLanguage(_this._options.lang).stringTypeMapping;
            var typeBuilder = new TypeBuilder_1.TypeGraphBuilder(stringTypeMapping);
            if (_this.isInputJSONSchema) {
                immutable_1.Map(_this._allInputs.schemas).forEach(function (schema, name) {
                    typeBuilder.addTopLevel(name, JSONSchemaInput_1.schemaToType(typeBuilder, name, schema));
                });
                var graph = typeBuilder.finish();
                GatherNames_1.gatherNames(graph);
                return graph;
            }
            else if (_this.isInputGraphQL) {
                var numInputs_1 = Object.keys(_this._allInputs.graphQLs).length;
                immutable_1.Map(_this._allInputs.graphQLs).forEach(function (_a, name) {
                    var schema = _a.schema, query = _a.query;
                    var newTopLevels = GraphQL_1.makeGraphQLQueryTypes(name, typeBuilder, schema, query);
                    newTopLevels.forEach(function (t, actualName) {
                        typeBuilder.addTopLevel(numInputs_1 === 1 ? name : actualName, t);
                    });
                });
                return typeBuilder.finish();
            }
            else {
                var doInferMaps = _this._options.inferMaps;
                var doInferEnums = _this._options.inferEnums;
                var doCombineClasses = _this._options.combineClasses;
                var samplesMap = immutable_1.Map(_this._allInputs.samples);
                var inputs = immutable_1.List([
                    doInferMaps,
                    doInferEnums,
                    doCombineClasses,
                    samplesMap.map(function (values) { return immutable_1.List(values); }),
                    _this._compressedJSON
                ]);
                var inputHash = undefined;
                if (_this._doCache) {
                    inputHash = inputs.hashCode();
                    var maybeGraph = graphByInputHash.get(inputHash);
                    if (maybeGraph !== undefined) {
                        return maybeGraph;
                    }
                }
                var inference_1 = new Inference_1.TypeInference(typeBuilder, doInferEnums);
                immutable_1.Map(_this._allInputs.samples).forEach(function (cjson, name) {
                    typeBuilder.addTopLevel(name, inference_1.inferType(_this._compressedJSON, name, false, cjson));
                });
                var graph = typeBuilder.finish();
                if (doCombineClasses) {
                    graph = CombineClasses_1.combineClasses(graph, stringTypeMapping);
                }
                if (doInferMaps) {
                    graph = InferMaps_1.inferMaps(graph, stringTypeMapping);
                }
                GatherNames_1.gatherNames(graph);
                if (inputHash !== undefined) {
                    graphByInputHash = graphByInputHash.set(inputHash, graph);
                }
                return graph;
            }
        };
        this.readSampleFromStream = function (name, readStream) { return __awaiter(_this, void 0, void 0, function () {
            var input, _a, _b, input;
            return __generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        if (!this.isInputJSONSchema) return [3 /*break*/, 2];
                        _b = (_a = JSON).parse;
                        return [4 /*yield*/, get_stream_1.getStream(readStream)];
                    case 1:
                        input = _b.apply(_a, [_c.sent()]);
                        if (_.has(this._allInputs.schemas, name)) {
                            throw new Error("More than one schema given for " + name);
                        }
                        this._allInputs.schemas[name] = input;
                        return [3 /*break*/, 4];
                    case 2: return [4 /*yield*/, this._compressedJSON.readFromStream(readStream)];
                    case 3:
                        input = _c.sent();
                        if (!_.has(this._allInputs.samples, name)) {
                            this._allInputs.samples[name] = [];
                        }
                        this._allInputs.samples[name].push(input);
                        _c.label = 4;
                    case 4: return [2 /*return*/];
                }
            });
        }); };
        this.run = function () { return __awaiter(_this, void 0, void 0, function () {
            var targetLanguage, _i, _a, source, name, schema, query, _b, _c, source, _d, _e, sample, _f, _g, _h, name, schema, graph;
            return __generator(this, function (_j) {
                switch (_j.label) {
                    case 0:
                        targetLanguage = getTargetLanguage(this._options.lang);
                        if (!isGraphQLData(this._options.sources)) return [3 /*break*/, 1];
                        for (_i = 0, _a = this._options.sources; _i < _a.length; _i++) {
                            source = _a[_i];
                            name = source.name, schema = source.schema, query = source.query;
                            this._allInputs.graphQLs[name] = { schema: schema, query: query };
                        }
                        return [3 /*break*/, 12];
                    case 1:
                        if (!isSourceData(this._options.sources)) return [3 /*break*/, 8];
                        _b = 0, _c = this._options.sources;
                        _j.label = 2;
                    case 2:
                        if (!(_b < _c.length)) return [3 /*break*/, 7];
                        source = _c[_b];
                        _d = 0, _e = source.samples;
                        _j.label = 3;
                    case 3:
                        if (!(_d < _e.length)) return [3 /*break*/, 6];
                        sample = _e[_d];
                        return [4 /*yield*/, this.readSampleFromStream(source.name, toReadable(sample.source))];
                    case 4:
                        _j.sent();
                        _j.label = 5;
                    case 5:
                        _d++;
                        return [3 /*break*/, 3];
                    case 6:
                        _b++;
                        return [3 /*break*/, 2];
                    case 7: return [3 /*break*/, 12];
                    case 8:
                        if (!isSchemaData(this._options.sources)) return [3 /*break*/, 12];
                        _f = 0, _g = this._options.sources;
                        _j.label = 9;
                    case 9:
                        if (!(_f < _g.length)) return [3 /*break*/, 12];
                        _h = _g[_f], name = _h.name, schema = _h.schema;
                        return [4 /*yield*/, this.readSampleFromStream(name, toReadable(schema))];
                    case 10:
                        _j.sent();
                        _j.label = 11;
                    case 11:
                        _f++;
                        return [3 /*break*/, 9];
                    case 12:
                        graph = this.makeGraph();
                        if (this._options.noRender) {
                            return [2 /*return*/, { lines: ["Done.", ""], annotations: immutable_1.List() }];
                        }
                        return [2 /*return*/, targetLanguage.renderGraphAndSerialize(graph, this._options.alphabetizeProperties, this._options.rendererOptions)];
                }
            });
        }); };
        this._options = _.assign(_.clone(defaultOptions), options);
        this._allInputs = { samples: {}, schemas: {}, graphQLs: {} };
        var mapping = getTargetLanguage(this._options.lang).stringTypeMapping;
        var makeDate = mapping.date !== "string";
        var makeTime = mapping.time !== "string";
        var makeDateTime = mapping.dateTime !== "string";
        this._compressedJSON = new CompressedJSON_1.CompressedJSON(makeDate, makeTime, makeDateTime);
    }
    Object.defineProperty(Run.prototype, "isInputJSONSchema", {
        get: function () {
            return isSchemaData(this._options.sources);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Run.prototype, "isInputGraphQL", {
        get: function () {
            return isGraphQLData(this._options.sources);
        },
        enumerable: true,
        configurable: true
    });
    return Run;
}());
exports.Run = Run;
function quicktype(options, useCache) {
    if (useCache === void 0) { useCache = true; }
    return new Run(options, useCache).run();
}
exports.quicktype = quicktype;
